__all__ = ()

CO_OPTIMIZED = 1
CO_NEWLOCALS = 2
CO_VARARGS = 4
CO_VARKEYWORDS = 8
CO_NESTED = 16
CO_GENERATOR = 32
CO_NOFREE = 64
    
CO_COROUTINE = 128
CO_ITERABLE_COROUTINE = 256
CO_ASYNC_GENERATOR = 512
# matches `async def` functions and `@coroutine` functions.
CO_COROUTINE_ALL = CO_COROUTINE | CO_ITERABLE_COROUTINE



__all__ = ()

import sys

from .compilation import maybe_compile
from .editor_base import _validate_buffer, EditorBase

from ....utils import copy_docs


class EditorSimple(EditorBase):
    """
    Simple `input` based editor.
    
    Attributes
    ----------
    buffer : `list` of `str`
        A precreated buffer which commands the initial lines of the input to write.
    compiled_code : `None`, ``CodeType``
        The compiled inputted code.
    file_name : `str`
        File name of the code produced by the editor.
    prefix_initial : `str`
        First line's prefix.
    prefix_continuous : `str`
        Non-first line's prefix.
    prefix_length : `int`
        As how long should the prefix's length be interpreted.
    """
    __slots__ = ('buffer', )
    
    @copy_docs(EditorBase.__new__)
    def __new__(cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history):
        buffer = _validate_buffer(buffer)
        
        self = EditorBase.__new__(
            cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history
        )
        self.buffer = buffer
        return self
    
    
    @copy_docs(EditorBase.__call__)
    def __call__(self):
        try:
            import readline
        except ImportError:
            pass
        
        buffer = self.buffer
        
        while True:
            if buffer:
                prefix = self.prefix_continuous
            else:
                prefix = self.prefix_initial
            
            try:
                line = input(prefix)  
            except EOFError as err:
                sys.stdout.write('\n')
                raise SystemExit() from err
            
            if not buffer:
                line = line.rstrip()
                if not line:
                    continue
            
            buffer.append(line)
            
            compiled_code = maybe_compile(buffer, self.file_name)
            if compiled_code is None:
                continue
            
            self.compiled_code = compiled_code
            break
        
        return self.compiled_code
    
    
    @copy_docs(EditorBase.get_buffer)
    def get_buffer(self):
        return self.buffer



from .compressors import *
from .cookiejar import *
from .exceptions import *
from .formdata import *
from .header_building_and_parsing import *
from .headers import *
from .helpers import *
from .http_message import *
from .http_protocol import *
from .http_stream_writer import *
from .mime_type import *
from .multipart import *
from .quoting import *
from .url import *
from .websocket_frame import *

from .import headers

__all__ = (
    'headers',
    *compressors.__all__,
    *cookiejar.__all__,
    *exceptions.__all__,
    *formdata.__all__,
    *header_building_and_parsing.__all__,
    *headers.__all__,
    *helpers.__all__,
    *http_message.__all__,
    *http_protocol.__all__,
    *http_stream_writer.__all__,
    *mime_type.__all__,
    *multipart.__all__,
    *quoting.__all__,
    *url.__all__,
    *websocket_frame.__all__,
)



__all__ = ()

from .docs import has_docs
from .utils import is_iterable

@has_docs
class DictionaryKeyIteratorBase:
    """
    Key iterator of a dictionary..
    
    Attributes
    ----------
    _parent : `dict`
        The parent dictionary.
    """
    __slots__ = ('_parent',)
    
    @has_docs
    def __init__(self, parent):
        """
        Creates a new dictionary key iterator bound to the given dictionary.
        
        Parameters
        ----------
        parent : `dict`
            The parent dictionary.
        """
        self._parent = parent
    
    
    @has_docs
    def __iter__(self):
        """
        Iterates over the dictionary's keys.
        
        This method is an iterable generator.
        
        Yields
        ------
        key : `Any`
        """
        return
        yield
    
    
    @has_docs
    def __contains__(self, contains_key):
        """Returns whether the respective dictionary contains the given key."""
        return False
    
    
    @has_docs
    def __len__(self):
        """Returns the respective dictionary's length."""
        return len(self._parent)
    
    
    @has_docs
    def __repr__(self):
        """Returns the dictionary key iterator's representation."""
        return f'<{self.__class__.__name__} to {self._parent!r}>'


    @has_docs
    def __eq__(self, other):
        """Returns whether the two dictionary key iterators are the same."""
        if isinstance(other, type(self)):
            return self._parent == other._parent
        
        return self._parent.__eq__(other)


@has_docs
class DictionaryValueIteratorBase:
    """
    Value iterator of a dictionary..
    
    Attributes
    ----------
    _parent : `dict`
        The parent dictionary.
    """
    __slots__ = ('_parent',)
    
    @has_docs
    def __init__(self, parent):
        """
        Creates a new dictionary value iterator bound to the given dictionary.
        
        Parameters
        ----------
        parent : `dict`
            The parent dictionary.
        """
        self._parent = parent
    
    
    @has_docs
    def __iter__(self):
        """
        Iterates over the dictionary's values.
        
        This method is an iterable generator.
        
        Yields
        ------
        value : `Any`
        """
        return
        yield
    
    
    @has_docs
    def __contains__(self, contains_value):
        """Returns whether the respective dictionary contains the given value."""
        return False
    
    
    @has_docs
    def __len__(self):
        """Returns the respective dictionary's length."""
        return len(self._parent)
    
    
    @has_docs
    def __repr__(self):
        """Returns the dictionary value iterator's representation."""
        return f'<{self.__class__.__name__} to {self._parent!r}>'
    
    
    @has_docs
    def __eq__(self, other):
        """Returns whether the two dictionary item iterators are the same."""
        if isinstance(other, type(self)):
            return self._parent == other._parent
        
        elif isinstance(other, list):
            if len(self) != len(other):
                return False
            
            other = other.copy()
        
        elif is_iterable(other):
            has_length_method = hasattr(type(other), '__len__')
            
            if has_length_method:
                if len(self) != len(other):
                    return False
            
            other = list(other)
            
            if not has_length_method:
                if len(self) != len(other):
                    return False
        
        else:
            return NotImplemented
        
        for value in self:
            try:
                other.remove(value)
            except ValueError:
                return False
        
        if other:
            return False
        
        return True


@has_docs
class DictionaryItemIteratorBase:
    """
    Item iterator of a dictionary..
    
    Attributes
    ----------
    _parent : `dict`
        The parent dictionary.
    """
    __slots__ = ('_parent',)
    
    @has_docs
    def __init__(self, parent):
        """
        Creates a new dictionary item iterator bound to the given dictionary.
        
        Parameters
        ----------
        parent : `dict`
            The parent dictionary.
        """
        self._parent = parent
    
    
    @has_docs
    def __iter__(self):
        """
        Iterates over the dictionary's items.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : `tuple` (`Any`, `str`)
        """
        return
        yield
    
    
    @has_docs
    def __contains__(self, contains_item):
        """Returns whether the respective dictionary contains the given item."""
        return False
    
    
    @has_docs
    def __len__(self):
        """Returns the respective dictionary's length."""
        return len(self._parent)
    
    
    @has_docs
    def __repr__(self):
        """Returns the dictionary item iterator's representation."""
        return f'<{self.__class__.__name__} to {self._parent!r}>'
    
    
    @has_docs
    def __eq__(self, other):
        """Returns whether the two dictionary item iterators are the same."""
        if isinstance(other, type(self)):
            return self._parent == other._parent
        
        elif isinstance(other, list):
            if len(self) != len(other):
                return False
            
            other = other.copy()
        
        elif is_iterable(other):
            has_length_method = hasattr(type(other), '__len__')
            
            if has_length_method:
                if len(self) != len(other):
                    return False
            
            other = list(other)
            
            if not has_length_method:
                if len(self) != len(other):
                    return False
        
        else:
            return NotImplemented
        
        for item in self:
            try:
                other.remove(item)
            except ValueError:
                return False
        
        if other:
            return False
        
        return True



__all__ = ()


class History:
    """
    Represents an editor's history.
    
    Attributes
    ----------
    _cursor_index : `int`
        The position where the cache lookup is at.
    _elements : `dict` of (`int`, `list` of `str`) items
        The added elements to the history.
    _line_cache : `dict` of (`str`, `str`) items
        Line cache caching same lines.
    _size : `int`
        The last registered content's index.
    """
    __slots__ = ('_cursor_index', '_elements', '_line_cache', '_size')
    
    def __new__(cls):
        """
        Creates a editor history.
        """
        self = object.__new__(cls)
        self._line_cache = {}
        self._cursor_index = 0
        self._elements = {}
        self._size = 0
        return self
    
    
    def __repr__(self):
        """Returns representation of the history."""    
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' cache: ')
        repr_parts.append(repr(len(self._line_cache)))
        
        repr_parts.append(', elements: ')
        repr_parts.append(repr(len(self._elements)))
        
        repr_parts.append(', size: ')
        repr_parts.append(repr(self._size))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def maybe_add_buffer_of(self, index, editor):
        """
        Adds the given editor's buffer into the history if it supports lookup.
        
        Parameters
        ----------
        editor : ``EditorBase`
            The editor to add it's content.
        buffer : `list` of `str`
            The buffer to add.
        
        Returns
        -------
        added : `bool`
            Whether anything was added.
        """
        self._size = index
        self._cursor_index = index + 1
        
        if (editor is None) or (not editor.has_history_support()):
            return False
        
        return self._add(index, editor.get_stripped_buffer())
    
    
    def _add(self, index, buffer):
        """
        Adds a new element into the history.
        
        Parameters
        ----------
        index : `int`
            The index of the element.
        buffer : `list` of `str`
            The buffer to add.
        
        Returns
        -------
        added : `bool`
            Whether anything was added.
        """
        if buffer:
            line_cache = self._line_cache
            
            element = [line_cache.setdefault(line, line) for line in buffer]
            self._elements[index] = element
            added = True
        
        else:
            added = False
        
        return added
    
    
    def reset_cursor_index(self):
        """
        Resets the cursor's position.
        """
        self._cursor_index = self._size + 1
    
    
    def get_at(self, index):
        """
        Returns the element at the specified position.
        
        Parameters
        ----------
        index : `list` of `int`
            The parameter's index to get.
        
        Returns
        -------
        element : `None`, `list` of `str`
        """
        return self._elements.get(index, None)
    
    
    def get_at_cursor_index_position(self):
        """
        Returns the element at the cursor's current position.
        
        Returns
        -------
        element : `None`, `list` of `str`
        """
        return self._elements.get(self._cursor_index, None)
    
    
    def get_previous(self):
        """
        Gets the previous element of the history moving the cursor.
        
        Returns
        -------
        previous : `None`, `list` of `str`
        """
        cursor_index = self._cursor_index
        elements = self._elements
        current = elements.get(cursor_index, None)
        
        while cursor_index > -1:
            cursor_index -= 1
            element = elements.get(cursor_index, None)
            if (element is not None) and element != current:
                previous = element
                break
        else:
            previous = None
        
        self._cursor_index = cursor_index
        return previous
    
    
    def get_next(self):
        """
        Gets the next element of the history moving the cursor.
        
        Returns
        -------
        next_ : `None`, `list` of `str`
        """
        cursor_index = self._cursor_index
        elements = self._elements
        current = elements.get(cursor_index, None)
        size = self._size
        
        while cursor_index < size:
            cursor_index += 1
            element = elements.get(cursor_index, None)
            if (element is not None) and element != current:
                next_ = element
                break
        else:
            next_ = None
        
        self._cursor_index = cursor_index
        return next_



__all__ = ('UnixReadPipeTransportLayer', 'UnixWritePipeTransportLayer')

import errno, os, sys
from stat import S_ISCHR, S_ISFIFO, S_ISSOCK

from ...utils import copy_docs, include

from ..traps import skip_ready_cycle

from .extra_info import EXTRA_INFO_NAME_PIPE, set_extra_info
from .transport_layer import TransportLayerBase


write_exception_async = include('write_exception_async')

MAX_READ_SIZE = 262144
IS_AIX = sys.platform.startswith('aix')


class UnixReadPipeTransportLayer(TransportLayerBase):
    """
    Asynchronous read only transport implementation for pipes.
    
    Attributes
    ----------
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The respective event loop of the transport.
    _closing : `bool`
        Whether the transport ic closing.
    _file_descriptor : `int`
        The used socket's file descriptor number.
    _pipe : `None`, `file-like`
        The pipe to connect to on read end.
        
        Is set to non-blocking mode.
        
        After closing the transport is set to `None`.
    _protocol : `None`, ``SubprocessReadPipeProtocol``
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    """
    __slots__ = ('_paused', '_closing', '_file_descriptor', '_pipe', '_protocol')
    
    async def __new__(cls, loop, extra, pipe, protocol):
        """
        Creates a new ``UnixReadPipeTransportLayer`` with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the transport.
        pipe : `file-like`
            The pipe to connect to on read end.
        protocol : ``SubprocessReadPipeProtocol``, `Any`
            Asynchronous protocol implementation used by the transport.
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        
        Raises
        ------
        ValueError
            If `pipe` was not given neither as pipe, socket or character device.
        """
        file_descriptor = pipe.fileno()
        mode = os.fstat(file_descriptor).st_mode
        if not (S_ISFIFO(mode) or S_ISSOCK(mode) or S_ISCHR(mode)):
            raise ValueError(
                f'`{cls.__name__}` is only for pipes, sockets and character devices, got '
                f'{pipe.__class__.__name__}; {pipe!r}.'
            )
        
        extra = set_extra_info(extra, EXTRA_INFO_NAME_PIPE, pipe)
        
        self = TransportLayerBase.__new__(cls, loop, extra)
        self._pipe = pipe
        self._file_descriptor = file_descriptor
        self._protocol = protocol
        self._closing = False
        self._paused = False
        
        try:
            os.set_blocking(file_descriptor, False)
            # skip 1 loop
            await skip_ready_cycle()
            
            protocol.connection_made(self)
            loop.add_reader(file_descriptor, self._read_ready)
        
        except:
            self.close()
            raise
        
        return self
    
    def __repr__(self):
        """Returns the transport layer's representation."""
        return f'<{self.__class__.__name__} file_descriptor={self._file_descriptor}>'
    
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        """
        try:
            data = os.read(self._file_descriptor, MAX_READ_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as err:
            self._fatal_error(err, 'Fatal read error on pipe transport')
        else:
            if data:
                self._protocol.data_received(data)
            else:
                self._closing = True
                loop = self._loop
                loop.remove_reader(self._file_descriptor)
                protocol = self._protocol
                loop.call_soon(protocol.__class__.eof_received, protocol)
                loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    
    @copy_docs(TransportLayerBase.pause_reading)
    def pause_reading(self):
        if self._closing or self._paused:
            return False
        
        self._paused = True
        self._loop.remove_reader(self._file_descriptor)
        return True
    
    
    @copy_docs(TransportLayerBase.resume_reading)
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        
        Returns
        -------
        reading_resume : `bool`
            Whether reading was resumed.
        """
        if self._closing or not self._paused:
            return False
        
        self._paused = False
        self._loop.add_reader(self._file_descriptor, self._read_ready)
        return True
    
    
    @copy_docs(TransportLayerBase.set_protocol)
    def set_protocol(self, protocol):
        self._protocol = protocol
    
    
    @copy_docs(TransportLayerBase.get_protocol)
    def get_protocol(self):
        return self._protocol
    
    
    @copy_docs(TransportLayerBase.is_closing)
    def is_closing(self):
        return self._closing
    
    
    @copy_docs(TransportLayerBase.close)
    def close(self):
        if not self._closing:
            self._close(None)
    
    
    def __del__(self):
        """
        Closes the read pipe transport if not yet closed.
        """
        pipe = self._pipe
        if (pipe is not None):
            pipe.close()
    
    
    @copy_docs(TransportLayerBase._fatal_error)
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        if not (isinstance(exception, OSError) and (exception.errno == errno.EIO)):
            write_exception_async(
                exception,
                [
                    message,
                    '\nException occurred at \n',
                    repr(self),
                    '.\n',
                ],
                loop = self._loop
            )
        
        self._close(exception)
    
    
    def _close(self, exception):
        """
        Starts the transport's closing process.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self._closing = True
        loop = self._loop
        loop.remove_reader(self._file_descriptor)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    
    def _call_connection_lost(self, exception):
        """
        Calls the read pipe transport's protocol's `.connection_lost` with the given exception and closes the
        transport's pipe.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Exception to call the protocol's ``.connection_lost`` with.
            
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        protocol = self._protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self._pipe
            if (pipe is not None):
                self._pipe = None
                pipe.close()
            
            self._protocol = None


class UnixWritePipeTransportLayer(TransportLayerBase):
    """
    Asynchronous write only transport implementation for pipes.
    
    Attributes
    ----------
    _buffer : `bytearray`
        Data ensured to be written on the wrapped pipe as it becomes readable again.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _closing : `bool`
        Whether the transport ic closing.
    _file_descriptor : `int`
        The used socket's file descriptor number.
    _loop : ``EventThread``
        The respective event loop of the transport.
    _pipe : `None`, `file-like`
        The pipe to connect to on read end.
        
        Is set to non-blocking mode.
        
        After closing the transport is set to `None`.
    _protocol : `None`, ``SubprocessWritePipeProtocol``, `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    _protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    """
    __slots__ = (
        '_buffer', '_extra', '_high_water', '_low_water', '_closing', '_file_descriptor', '_loop', '_pipe',
        '_protocol', '_protocol_paused'
    )
    
    async def __new__(cls, loop, pipe, protocol, extra=None):
        """
        Creates a new ``UnixWritePipeTransportLayer`` with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the transport.
        pipe `: file-like` object
            The pipe to connect to on read end.
        protocol : ``SubprocessWritePipeProtocol``, `Any`
            Asynchronous protocol implementation used by the transport.
        extra : `None`, `dict` of (`str`, `Any`) items = `None`, Optional
            Optional transport information.

        Raises
        ------
        ValueError
            If `pipe` was not given neither as pipe, socket or character device.
        """
        file_descriptor = pipe.fileno()
        mode = os.fstat(file_descriptor).st_mode
        is_char = S_ISCHR(mode)
        is_fifo = S_ISFIFO(mode)
        is_socket = S_ISSOCK(mode)
        if not (is_char or is_fifo or is_socket):
            raise ValueError(
                f'{cls.__name__} is only for pipes, sockets and character devices, got '
                f'{pipe.__class__.__name__}; {pipe!r}.'
            )
        

        extra = set_extra_info(extra, EXTRA_INFO_NAME_PIPE, pipe)
        
        self = TransportLayerBase.__new__(cls, loop, extra)
        self._extra = extra
        self._loop = loop
        self._protocol_paused = False
        self._pipe = pipe
        self._file_descriptor = file_descriptor
        self._protocol = protocol
        self._buffer = bytearray()
        self._closing = False  # Set when close() or write_eof() called.
        
        self._high_water = 65536
        self._low_water = 16384
        
        try:
            os.set_blocking(file_descriptor, False)
            # skip 1 loop
            await skip_ready_cycle()
            
            protocol.connection_made(self)
            
            # On AIX, the reader trick (to be notified when the read end of the  socket is closed) only works for
            # sockets. On other platforms it works for pipes and sockets.
            if is_socket or (is_fifo and not IS_AIX):
                loop.add_reader(file_descriptor, self._read_ready)
        except:
            self.close()
            raise
        
        return self
    
    def __repr__(self):
        """Returns the transport layer's representation."""
        return f'<{self.__class__.__name__} file_descriptor={self._file_descriptor}>'
    
    
    @copy_docs(TransportLayerBase.get_write_buffer_size)
    def get_write_buffer_size(self):
        return len(self._buffer)
    
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        
        If this happens, since it is a write only pipe, means it should be closed, so we do like that.
        """
        # Pipe was closed by peer.
        if self._buffer:
            exception = BrokenPipeError()
        else:
            exception = None
            
        self._close(exception)
    
    
    @copy_docs(TransportLayerBase.write)
    def write(self, data):
        if not data:
            return
        
        if isinstance(data, bytearray):
            data = memoryview(data)
        
        if self._closing:
            return
        
        buffer = self._buffer
        if not buffer:
            try:
                n = os.write(self._file_descriptor, data)
            except (BlockingIOError, InterruptedError):
                n = 0
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on pipe transport')
                return
            
            if n == len(data):
                return
            
            if n > 0:
                data = memoryview(data)[n:]
            
            self._loop.add_writer(self._file_descriptor, self._write_ready)
        
        buffer.extend(data)
        self._maybe_pause_protocol()
    
    
    def _write_ready(self):
        """
        Added as a write callback on the respective event loop when the transport has unsent data. Called when the
        respective socket becomes writable.
        """
        buffer = self._buffer
        
        try:
            n = os.write(self._file_descriptor, buffer)
        except (BlockingIOError, InterruptedError):
            pass
        
        except BaseException as err:
            buffer.clear()
            self._loop.remove_writer(self._file_descriptor)
            self._fatal_error(err, 'Fatal write error on pipe transport')
        else:
            if n == len(buffer):
                buffer.clear()
                self._loop.remove_writer(self._file_descriptor)
                self._maybe_resume_protocol()  # May append to buffer.
                if self._closing:
                    self._loop.remove_reader(self._file_descriptor)
                    self._call_connection_lost(None)
                return
            
            if n > 0:
                del buffer[:n]
    
    
    @copy_docs(TransportLayerBase.can_write_eof)
    def can_write_eof(self):
        return True
    
    
    @copy_docs(TransportLayerBase.write_eof)
    def write_eof(self):
        if self._closing:
            return
        
        self._closing = True
        if not self._buffer:
            loop = self._loop
            loop.remove_reader(self._file_descriptor)
            loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    
    @copy_docs(TransportLayerBase.set_protocol)
    def set_protocol(self, protocol):
        self._protocol = protocol
    
    
    @copy_docs(TransportLayerBase.get_protocol)
    def get_protocol(self):
        return self._protocol
    
    
    @copy_docs(TransportLayerBase.get_protocol)
    def is_closing(self):
        return self._closing
    
    
    @copy_docs(TransportLayerBase.close)
    def close(self):
        if (self._pipe is not None) and (not self._closing):
            self.write_eof()
    
    def __del__(self):
        """
        Closes the write pipe transport if not yet closed.
        """
        pipe = self._pipe
        if (pipe is not None):
            pipe.close()
    
    
    @copy_docs(TransportLayerBase.abort)
    def abort(self):
        self._close(None)
    
    
    @copy_docs(TransportLayerBase._fatal_error)
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        if not isinstance(exception, OSError):
            write_exception_async(
                exception,
                [
                    message,
                    '\nException occurred at \n',
                    repr(self),
                    '.\n',
                ],
                loop = self._loop,
            )
        
        self._close(exception)
    
    
    def _close(self, exception):
        """
        Starts the transport's closing process.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self._closing = True
        
        loop = self._loop
        buffer = self._buffer
        if buffer:
            self._loop.remove_writer(self._file_descriptor)
            buffer.clear()
        
        loop.remove_reader(self._file_descriptor)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    
    def _call_connection_lost(self, exception):
        """
        Calls the write pipe transport's protocol's `.connection_lost` with the given exception and closes the
        transport's pipe.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Exception to call the protocol's ``.connection_lost`` with.
            
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        protocol = self._protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self._pipe
            if (pipe is not None):
                self._pipe = None
                pipe.close()
            
            self._protocol = None
    
    
    def _maybe_pause_protocol(self):
        """
        Called after data was ensured to be written into the pipe to check whether it's protocol should be paused.
        """
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self._protocol_paused:
            return
        
        self._protocol_paused = True
        
        protocol = self._protocol
        if protocol is None:
            return
        
        try:
            protocol.pause_writing()
        except BaseException as err:
            write_exception_async(
                err,
                [
                    repr(self),
                    '`._maybe_pause_protocol` failed\nOn: ',
                    repr(protocol),
                    '.pause_writing()\n'
                ],
                loop = self._loop,
            )
    
    
    def _maybe_resume_protocol(self):
        """
        Called after successful writing to the pipe to check whether the protocol should be resumed.
        """
        if (self._protocol_paused and self.get_write_buffer_size() <= self._low_water):
            self._protocol_paused = False
            protocol = self._protocol
            if (protocol is not None):
                try:
                    protocol.resume_writing()
                except BaseException as err:
                    write_exception_async(
                        err,
                        [
                            repr(self),
                            '`._maybe_resume_protocol` failed\nOn: ',
                            repr(protocol),
                            '.resume_writing()\n'
                        ],
                        loop = self._loop,
                    )
    
    
    @copy_docs(TransportLayerBase.get_write_buffer_limits)
    def get_write_buffer_limits(self):
        return self._low_water, self._high_water
    
    
    def _set_write_buffer_limits(self, low=None, high=None):
        """
        Sets the write buffer limits of the transport.
        
        Parameters
        ----------
        low : None`, `int` = `None`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None`, `int` = `None`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.

        Raises
        ------
        ValueError
            If `high` is lower than `low` or if `low` is lower than `0`.
        """
        if high is None:
            if low is None:
                high = 65536
                low = 16384
            else:
                high = low << 2
        else:
            if low is None:
                low = high >> 2
        
        if low < 0 or high < low:
            raise ValueError(
                f'High water must be greater or equal than low, what must be greater than equal than `0`, '
                f'got high={high!r}; low={low!r}.'
            )
        
        self._high_water = high
        self._low_water = low
    
    
    @copy_docs(TransportLayerBase.set_write_buffer_limits)
    def set_write_buffer_limits(self, low=None, high=None):
        self._set_write_buffer_limits(low=low, high=high)
        self._maybe_pause_protocol()



__version__ = '1.0.41'

from .core import *
from .ext import *
from .utils import *

__all__ = (
    *core.__all__,
    *ext.__all__,
    *utils.__all__,
)

from .utils.export_include import check_satisfaction
check_satisfaction()



__all__ = ('RemovedDescriptor', )

from .docs import has_docs


@has_docs
class RemovedDescriptor:
    """
    A descriptor, what can be used to overwrite a class's attribute, what should be inherited anyways.
    
    Attributes
    ----------
    name : `None`, `str`
        The name of the attribute. Set when the class is finalizing.
    """
    __slots__ = ('name',)
    def __init__(self,):
        self.name = None
    
    def __set_name__(self, owner, name):
        self.name = name
    
    def __get__(self, obj, type_):
        name = self.name
        if name is None:
            raise RuntimeError(f'`{self.__class__.__name__}` is not initialized correctly yet.')
        
        if obj is None:
            error_message = f'Type object `{type_.__name__!r}` has no attribute {name!r}'
        else:
            error_message = f'`{obj.__class__.__name__}` object has no attribute {name!r}'
        
        raise AttributeError(error_message)
    
    def __set__(self, obj, value):
        name = self.name
        if name is None:
            raise RuntimeError(f'`{self.__class__.__name__}` is not initialized correctly yet.')
        
        raise AttributeError(name)
    
    def __delete__(self, obj):
        name = self.name
        if name is None:
            raise RuntimeError(f'`{self.__class__.__name__}` is not initialized correctly yet.')
        
        raise AttributeError(name)



__all__ = ('Gatherer',)

from ...utils import ignore_frame, is_awaitable

from ..exceptions import InvalidStateError

from .future import FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING
from .result_gathering_future import ResultGatheringFuture


ignore_frame(__spec__.origin, '__call__', 'raise exception',)
ignore_frame(__spec__.origin, '__call__', 'future.get_result()',)

class GathererElement:
    """
    An element of a ``Gatherer`` results.
    
    Attributes
    ----------
    result : `Any`
        The result of a gathered future.
    exception : `None`, `BaseException`
        The exception of a gathered exception.
    """
    __slots__ = ('exception', 'result',)
    
    def __init__(self, result, exception):
        """
        Creates a new ``GathererElement`` with the given parameters.
        
        Parameters
        ----------
        result : `Any`
            The result of a gathered future.
        exception : `None`, `BaseException`
            The exception of a gathered exception.
        """
        self.result = result
        self.exception = exception
    
    
    def __call__(self):
        """
        Returns a gathered future's result or raises it's exception.
        
        Returns
        -------
        result : `Any`
            A gathered future's result.
        
        Raises
        ------
        BaseException
            A gathered future's exception.
        """
        exception = self.exception
        if exception is None:
            return self.result
        raise exception


class GathererCallback:
    """
    Callback of ``Gatherer`` to set a waited future's result or exception to itself.
    
    Attributes
    ----------
    _parent : ``Gatherer``
        The gathering future
    """
    __slots__ = ('_parent',)
    
    def __init__(self, parent):
        """
        Creates a new gathering future callback.
        
        Parameters
        ----------
        parent : ``Gatherer``
            The gathering future
        """
        self._parent = parent
    
    def __call__(self, future):
        """
        Sets a result of an exception to the parent gatherer future.
        
        Parameters
        ----------
        future : ``Future``
            A waited future.
        """
        parent = self._parent
        try:
            result = future.get_result()
        except BaseException as err:
            parent.set_exception_if_pending(err)
        else:
            parent.set_result_if_pending(result)


class Gatherer(ResultGatheringFuture):
    """
    A ``Future`` subclass to gather more future's results and their exceptions.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _loop : ``EventThread``
        The loop to what the created future is bound.
    _result : `list` of ``GathererElement``
        The already gathered future's results.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used only if
        `__debug__` is set as `True`.
    _count : `int`
        The amount, for how much future's result the gatherer waits.
    """
    __slots__ = ()
    
    def __new__(cls, loop, coroutines_or_futures):
        """
        Creates a new gatherer bound to the given `loop`, waiting for the given `coroutines_or_futures`-s results.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the created future will be bound to.
        coroutines_or_futures : `iterable` of `awaitable`
            Awaitables, which result will be gathered.
        
        Raises
        ------
        TypeError
            Any of the given `coroutines_or_futures` is not awaitable.
        """
        awaitables = set()
        for awaitable in coroutines_or_futures:
            if not is_awaitable(awaitable):
                raise TypeError(
                    f'Cannot await on {awaitable.__class__.__name__}; {awaitable!r}.'
                )
            awaitables.add(awaitable)
        
        self = object.__new__(cls)
        self._loop = loop
        self._count = len(awaitables)
        
        self._result = []
        self._exception = None
        
        self._callbacks = []
        self._blocking = False
        
        if awaitables:
            self._state = FUTURE_STATE_PENDING
            
            callback = GathererCallback(self)
            
            for awaitable in awaitables:
                task = loop.ensure_future(awaitable)
                task.add_done_callback(callback)
        else:
            self._state = FUTURE_STATE_FINISHED
        
        return self
    
    
    def set_result(self, result):
        """
        Sets a result to the gatherer. if all the expected results of the gatherer are retrieved already, marks it as
        done as well.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Raises
        ------
        InvalidStateError
            If the gatherer is already done.
        """
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_result')
        
        results = self._result
        results.append(GathererElement(result, None))
        if self._count != len(results):
            return
        
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
    
    
    def set_result_if_pending(self, result):
        """
        Sets a result to the gatherer. If all the expected results of the gatherer are retrieved already, marks it as
        done as well. Not like ``.set_result``, this method will not raise ``InvalidStateError`` if the future is
        already done.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Returns
        ------
        set_result : `int` (`0`, `1`, `2`)
            If the gatherer is already done, returns `0`. If the gatherer's result was successfully set, returns `1`,
            meanwhile if the gatherer was marked as done as well, returns `2`.
        """
        if self._state != FUTURE_STATE_PENDING:
            return 0
        
        results = self._result
        results.append(GathererElement(result, None))
        if self._count != len(results):
            return 2
        
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
        return 1
    
    
    def set_exception(self, exception):
        """
        Sets an exception to the gatherer. If all the expected results of the gatherer are retrieved already, marks it
        as done as well.
        
        Parameters
        ----------
        exception : `BaseException`
            The exception to set as a result of the gatherer.
        
        Raises
        ------
        InvalidStateError
            If the gatherer is already done.
        TypeError
            If `StopIteration` is given as `exception`.
        """
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_exception')
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        results = self._result
        results.append(GathererElement(None, exception))
        if self._count != len(results):
            return
        
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
    
    
    def set_exception_if_pending(self, exception):
        """
        Sets an exception to the gatherer. If all the expected results of the gatherer are retrieved already, marks it
        as done as well. Not like ``.set_exception``, this method will not raise ``InvalidStateError`` if the future is
        already done.
        
        Parameters
        ----------
        exception : `BaseException`
            The exception to set as a result of the gatherer.
        
        Returns
        ------
        set_result : `int` (`0`, `1`, `2`)
            If the gatherer is already done, returns `0`. If the exception was set successfully as a gatherer result,
            returns `1`, meanwhile if the gatherer was marked as done as well, returns `2`.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        if self._state != FUTURE_STATE_PENDING:
            return 0
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        results = self._result
        results.append(GathererElement(None, exception))
        if self._count != len(results):
            return 1
        
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
        return 2



__all__ = ()

from hashlib import md5, sha1, sha256


HASH_FUNCTION_BY_DIGEST_LENGTH = {
    16: md5,
    20: sha1,
    32: sha256,
}

class Fingerprint:
    """
    HTTP fingerprinting can be used to automate information systems and security audits. Automated security testing
    tools can use HTTP fingerprinting to narrow down the set of tests required, based on the specific platform or the
    specific web server being audited.
    
    Attributes
    ----------
    fingerprint : `bytes`
        The fingerprint's value.
    hash_function : `function`
        Hash function used by the fingerprint.
    
    Class Attributes
    ----------------
    HASH_FUNCTION_BY_DIGEST_LENGTH : `dict` of (`int`, `function`) items
        `fingerprint`'s length - `hash-function` relation mapping.
    """
    __slots__ = ('fingerprint', 'hash_function',)
    
    def __new__(cls, fingerprint):
        """
        Creates a new ``Fingerprint`` with the given parameters.
        
        Parameters
        ----------
        fingerprint : `bytes`
            Fingerprint value.
        
        Raises
        ------
        ValueError
            - If `fingerprint`'s length is not any of the expected ones.
            - If the detected `hash_function` is `md5` / `sha1`.
        """
        fingerprint_length = len(fingerprint)
        
        try:
            hash_function = HASH_FUNCTION_BY_DIGEST_LENGTH[fingerprint_length]
        except KeyError:
            raise ValueError(
                f'`fingerprint` has invalid length, got {fingerprint_length!r}, {fingerprint!r}.'
            ) from None
        
        if hash_function is md5 or hash_function is sha1:
            raise ValueError(
                '`md5` and `sha1` are insecure and not supported, use `sha256`.'
            )
        
        self = object.__new__(cls)
        self.hash_function = hash_function
        self.fingerprint = fingerprint
        return self
    
    
    def check(self, protocol):
        """
        Checks whether the given protocol's ssl data matches the fingerprint.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous protocol implementation.
        
        Raises
        ------
        ValueError
            If the fingerprint don't match.
        """
        if protocol.get_extra_info('ssl_context') is None:
            return
        
        ssl_object = protocol.get_extra_info('ssl_object')
        cert = ssl_object.getpeercert(binary_form=True)
        received = self.hash_function(cert).digest()
        fingerprint = self.fingerprint
        if received == fingerprint:
            return
            
        host, port, *_ = protocol.get_extra_info('peer_name')
        
        raise ValueError(
            f'The expected fingerprint: {fingerprint!r} not matches the received; received={received!r}; '
            f'host={host!r}; port={port!r}.'
        )



__all__ = ('RequestContextManager', 'WebSocketContextManager')

class RequestContextManager:
    """
    Asynchronous context manager wrapping a request coroutine.
    
    Examples
    --------
    ``RequestContextManager`` are returned by ``HTTPClient`` request methods. Request context managers can be used as an
    asynchronous context manager or as a simple awaitable.
    
    ```py
    async with http_client.get('http://python.org') as response:
        data = await response.read()
    ```
    
    ```py
    response = await http_client.get('http://python.org')
    data = await response.read()
    ```
    
    Attributes
    ----------
    coroutine : `None`, `CoroutineType`
        The wrapped requester coroutine.
    response : `None`, ``ClientResponse``
        Received client response if applicable.
    """
    __slots__ = ('coroutine', 'response', )
    
    def __init__(self, coroutine):
        """
        Creates a new request content manager.
        
        Parameters
        ----------
        coroutine : `CoroutineType`
            Requester coroutine to wrap.
        """
        self.coroutine = coroutine
        self.response = None
    
    
    def __iter__(self):
        """
        Awaits the wrapped coroutine.
        
        This method is a generator. Should be used with `await` expression.
        
        Returns
        -------
        response : ``ClientResponse``
            Received client response if applicable.
        
        Raises
        ------
        BaseException
            Any exception raised by the request.
        """
        return (yield from self.__aenter__().__await__())
    
    
    __await__ = __iter__
    
    
    async def __aenter__(self):
        """
        Enters the ``RequestContextManager`` as an asynchronous context manager. Releases the response when the context manager is
        exited.
        
        This method is a coroutine.
        
        Returns
        -------
        response : ``ClientResponse``
            Received client response if applicable.
        
        Raises
        ------
        BaseException
            Any exception raised by the request.
        """
        coroutine = self.coroutine
        if (coroutine is None):
            response = self.response
        else:
            response = await coroutine
            self.response = response
            self.coroutine = None
        
        return response
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the response.
        
        This method is a coroutine.
        """
        response = self.response
        if (response is not None):
            self.response = None
            response.release()


class WebSocketContextManager:
    """
    Asynchronous context manager wrapping a websocket connecting coroutine.
    
    Examples
    --------
    ``WebSocketContextManager``-s are returned by the ``HTTPClient.connect_websocket`` method. WebSocket
    context managers can be used as an asynchronous context manager or as a simple awaitable.
    
    ```py
    async with http_client.connect_websocket('http://ayaya.aya') as websocket:
        await websocket.send('ayaya')
    ```
    
    ```py
    websocket = await http_client.connect_websocket('http://ayaya.aya')
    await websocket.send('ayaya')
    ```
    
    Attributes
    ----------
    coroutine : `CoroutineType`
        The wrapped requester coroutine.
    websocket : `None`, ``WebSocketClient``
        The connected websocket client if applicable
    """
    __slots__ = ('coroutine', 'websocket', )
    
    def __init__(self, coroutine):
        """
        Creates a new websocket content manager.
        
        Parameters
        ----------
        coroutine : `CoroutineType`
            WebSocket connecting coroutine to wrap.
        """
        self.coroutine = coroutine
        self.websocket = None
    
    
    def __iter__(self):
        """
        Awaits the wrapped coroutine.
        
        This method is a generator. Should be used with `await` expression.
        
        Returns
        -------
        websocket : ``WebSocketClient``
            The connected websocket client.
        
        Raises
        ------
        BaseException
            Any exception raised meanwhile connecting the websocket.
        """
        return (yield from self.__aenter__().__await__())
    
    __await__ = __iter__
    
    
    async def __aenter__(self):
        """
        Enters the ``WebSocketContextManager`` as an asynchronous context manager. Closes the websocket when the
        context manager is exited.
        
        This method is a coroutine.
        
        Returns
        -------
        websocket : ``WebSocketClient``
            The connected websocket client.
        
        Raises
        ------
        BaseException
            Any exception raised meanwhile connecting the websocket.
        """
        coroutine = self.coroutine
        if (coroutine is None):
            websocket = self.websocket
        else:
            websocket = await coroutine
            self.websocket = websocket
            self.coroutine = None
        
        return websocket
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Closes the connected websocket.
        
        This method is a coroutine.
        """
        websocket = self.websocket
        if (websocket is not None):
            self.websocket = None
            await websocket.close()



__all__ = ('Cycler',)

from threading import current_thread

from ...utils import CallableAnalyzer, include

from ..time import LOOP_TIME
from ..traps import Task


write_exception_async = include('write_exception_async')


class CyclerCallable:
    """
    An element of a ``Cycler``, which describes whether the stored callable is sync or async and what is it's ordering
    priority.
    
    Attributes
    ----------
    func : `callable`
        The function to call.
    is_async : `bool`
        Whether `func` is async.
    priority : `int`
        Call order priority inside of a ``Cycler``.
    """
    __slots__ = ('func', 'is_async', 'priority')
    
    def __new__(cls, func, priority):
        """
        Creates a new ``CyclerCallable`` with the given parameters.
        
        Parameters
        ----------
        func : `callable`
            The function to call.
        priority : `int`
            Call order priority inside of a ``Cycler``.
        
        Raises
        ------
        TypeError
            - `priority` is not `int`, neither other numeric convertible to it.
            - `func` is not `callable`.
            - `func` accepts less or more reserved positional parameters than `1`.
        """
        if type(func) is cls:
            return func
        
        priority_type = type(priority)
        if (priority_type is not int):
            try:
                __int__ = getattr(priority, '__int__')
            except AttributeError:
                raise TypeError(
                    f'`priority` can be `int`, `numeric`, got {priority_type.__name__}; {priority!r}.'
                ) from None
            
            priority = __int__(priority)
        
        analyzer = CallableAnalyzer(func)
        min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
        if min_ > 1:
            raise TypeError(
                f'`{func!r}` excepts at least `{min_!r}` non reserved parameters, meanwhile `1` would be '
                'passed to it.'
            )
        
        if not ((min_ == 1) or max_ >= 1 or analyzer.accepts_args()):
            raise TypeError(
                f'`{func!r}` expects maximum `{max_!r}` non reserved parameters, meanwhile the event '
                'expects to pass `1`.'
            )
        
        is_async = analyzer.is_async()
        
        self = object.__new__(cls)
        self.func = func
        self.is_async = is_async
        self.priority = priority
        
        return self
    
    def __repr__(self):
        """Returns the cycler callable's representation."""
        return f'{self.__class__.__name__}(func={self.func!r}, priority={self.priority!r})'
    
    def __gt__(self, other):
        """Returns whether the priority of this cycler callable is greater than the other's."""
        return self.priority > other.priority
    
    def __ge__(self, other):
        """
        Returns whether the priority of this cycler callable is greater than the other's, or if both instance is equal.
        """
        self_priority = self.priority
        other_priority = other.priority
        if self_priority > other_priority:
            return True
        
        if other_priority == other_priority:
            if self.func == other.func:
                return True
        
        return False
    
    def __eq__(self, other):
        """Returns whether this cycler callable equals to the other."""
        if self.priority != other.priority:
            return False
        
        if self.func != other.func:
            return False
        
        return True
    
    def __ne__(self, other):
        """Returns whether this cycler callable not equals to the other."""
        if self.priority != other.priority:
            return True
        
        if self.func != other.func:
            return True
        
        return False
    
    def __le__(self, other):
        """
        Returns whether the priority of this cycler callable is less than the other's, or if both instance is equal.
        """
        self_priority = self.priority
        other_priority = other.priority
        if self_priority < other_priority:
            return True
        
        if other_priority == other_priority:
            if self.func == other.func:
                return True
        
        return False

    def __lt__(self, other):
        """Returns whether the priority of this cycler callable is less than the other's."""
        return self.priority < other.priority



class Cycler:
    """
    Cycles the given functions on an event loop, by calling them after every `n` amount of seconds.
    
    Attributes
    ----------
    cycle_time : `float`
        The time interval of the cycler to call the added functions.
    funcs : `list` of ``CyclerCallable``
        Callables of a cycler containing whether they are sync, async, and what is their priority order.
    handle : ``TimerHandle``
        Handler, what will call the cycler when cycle time is over.
    loop : ``EventThread``
        The async event loop of the cycler, what it uses to ensure itself.
    """
    __slots__ = ('cycle_time', 'funcs', 'handle', 'loop',)
    
    def __new__(cls, loop, cycle_time, *funcs, priority=0):
        """
        Creates a new ``Cycler`` with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The async event loop of the cycler, what it uses to ensure itself.
        cycle_time : `float`
            The time interval of the cycler to call the added functions.
        *funcs : `callable`
            Callables, what the cycler will call.
        priority : `int` = `0`, Optional (Keyword only)
            Priority order of the added callables, which define in which order the given `funcs` will be called.
        
        Raises
        ------
        RuntimeError
            Event loop closed.
        TypeError
            - `cycle_time` is not `float`, neither other numeric convertible to it.
            - `priority` is not `int`, neither other numeric convertible to it.
            - Any `func` is not `callable`.
            - Any `func` accepts less or more reserved positional parameters than `1`.
        ValueError
            If `cycle_time` is negative or `0`.
        
        Notes
        -----
        If the respective event loop is not running yet, then creating a ``Cycler`` will not start it either.
        """
        if not loop.running and not loop.should_run:
            raise RuntimeError('Event loop is closed.')
        
        cycle_time_type = cycle_time.__class__
        if cycle_time_type is not float:
            try:
                __float__ = getattr(cycle_time_type, '__float__')
            except AttributeError:
                raise TypeError(
                    f'`cycle_time` can be `float`, `numeric`, got {cycle_time_type.__name__}; {cycle_time!r}.'
                ) from None
            
            cycle_time = __float__(cycle_time)
        
        if cycle_time <= 0.0:
            raise ValueError(
                f'{cycle_time} cannot be `0` or less, got {cycle_time!r}.'
            )
        
        priority_type = type(priority)
        if (priority_type is not int):
            try:
                __int__ = getattr(priority, '__int__')
            except AttributeError:
                raise TypeError(
                    f'`priority` can be `int`, `numeric`, got {priority_type.__name__}; {priority!r}.'
                ) from None
            
            priority = __int__(priority)
        
        validated_funcs = []
        
        if funcs:
            for func in funcs:
                validated_func = CyclerCallable(func, priority)
                validated_funcs.append(validated_func)
            
            validated_funcs.sort()
        
        self = object.__new__(cls)
        self.loop = loop
        self.funcs = validated_funcs
        self.cycle_time = cycle_time
        if current_thread() is loop:
            handle = loop.call_later(cycle_time, cls._run, self)
        else:
            handle = loop.call_soon_thread_safe_lazy(loop.__class__.call_later, loop, cycle_time, cls._run, self)
        
        self.handle = handle
        
        return self
    
    def _run(self):
        """
        Runs all the functions added to the cycler.
        """
        for func in self.funcs:
            try:
                result = func.func(self)
                if func.is_async:
                    Task(result, self.loop)
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        self.__class__.__name__,
                        ' exception occurred\nat calling ',
                        repr(func),
                        '\n',
                    ],
                    loop = self.loop,
                )
        
        self.handle = self.loop.call_later(self.cycle_time, self.__class__._run, self)
    
    def __repr__(self):
        """Returns the cycler's representation."""
        repr_parts = [
            self.__class__.__name__,
            '(',
            repr(self.loop),
            ', ',
            str(self.cycle_time)
        ]
        
        funcs = self.funcs
        limit = len(funcs)
        if limit:
            priority = funcs[0].priority
            
            index = 1
            while index < limit:
                func = funcs[index]
                if func.priority == priority:
                    index += 1
                    continue
                
                for func in funcs:
                    repr_parts.append(', ')
                    repr_parts.append(repr(func))
                break
            
            else:
                for func in funcs:
                    repr_parts.append(', ')
                    repr_parts.append(repr(func.func))
                
                repr_parts.append(', priority=')
                repr_parts.append(repr(priority))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    def cancel(self):
        """
        Cancels the cycler.
        
        If called from an other thread than it's event loop, then ensures ``._cancel`` on it instead of calling it
        right away.
        """
        loop = self.loop
        if current_thread() is loop:
            self._cancel()
            return
        
        loop.call_soon_thread_safe_lazy(self.__class__._cancel, self)
    
    
    def _cancel(self):
        """
        Cancels the cycler.
        
        This method always runs on the cycler's event loop.
        """
        handle = self.handle
        if (handle is not None):
            self.handle = None
            handle.cancel()
    
    
    def call_now(self):
        """
        Calls the cycler now, doing it's cycle.
        
        If called from an other thread than it's event loop, then ensures ``._call_now`` on it instead of calling it
        right away.
        """
        loop = self.loop
        if current_thread() is loop:
            self._call_now()
            return
        
        loop.call_soon_thread_safe_lazy(self.__class__._call_now, self)
    
    
    def _call_now(self):
        """
        Calls the cycler now, doing it's cycle.
        
        This method always runs on the cycler's event loop.
        """
        handle = self.handle
        if (handle is not None):
            handle.cancel()
        
        self._run()
    
    
    def reschedule(self):
        """
        Reschedules the cycler, making it's cycle to start since now. If the cycler is not running, also starts it.
        
        If called from an other thread than it's event loop, then ensures ``._reschedule`` on it instead of calling it
        right away.
        """
        loop = self.loop
        if current_thread() is loop:
            self._reschedule()
            return
        
        loop.call_soon_thread_safe_lazy(self.__class__._reschedule, self)
    
    
    def _reschedule(self):
        """
        Reschedules the cycler, making it's cycle to start since now. If the cycler is not running, also starts it.
        
        This method always runs on the cycler's event loop.
        """
        handle = self.handle
        if (handle is not None):
            handle.cancel()
        
        self.handle = self.loop.call_later(self.cycle_time, self.__class__._run, self)
    
    
    @property
    def running(self):
        """
        Returns whether the cycler is currently running.
        
        Returns
        -------
        running : `str`
        """
        return (self.handle is not None)
    
    
    def set_cycle_time(self, cycle_time):
        """
        Sets the cycle time of the cycler to the given value.
        
        Parameters
        ----------
        cycle_time : `float`
            The time interval of the cycler to call the added functions.
        
        Raises
        ------
        TypeError
            `cycle_time` is not `float`, neither other numeric convertible to it.
        ValueError
            If `cycle_time` is negative or `0`.
        """
        cycle_time_type = cycle_time.__class__
        if cycle_time_type is not float:
            try:
                __float__ = getattr(cycle_time_type, '__float__')
            except AttributeError:
                raise TypeError(
                    f'`cycle_time` can be `float`, `numeric`, got {cycle_time_type.__name__}; {cycle_time!r}.'
                ) from None
            
            cycle_time = __float__(cycle_time)
        
        if cycle_time <= 0.0:
            raise ValueError(
                f'{cycle_time} cannot be `0` or less, got {cycle_time!r}.'
            )
        
        self.cycle_time = cycle_time
    
    
    def append(self, func, priority=0):
        """
        Adds the given `func` to the cycler to call.
        
        If called from an other thread than it's event loop, then it will ensure adding the `func` on it's own.
        
        Parameters
        ----------
        func : `callable`
            Callable, what the cycler will call.
        priority : `int` = `0`, Optional
            Priority order of the added callables, which define in which order the given `funcs` will be called.
        
        Raises
        ------
        TypeError
            - `priority` is not `int`, neither other numeric convertible to it.
            - Any `func` is not `callable`.
            - Any `func` accepts less or more reserved positional parameters than `1`.
        """
        validated_func = CyclerCallable(func, priority)
        
        loop = self.loop
        if current_thread() is loop:
            self._append(validated_func)
            return
        
        loop.call_soon_thread_safe_lazy(self.__class__._append, self, validated_func)
    
    
    def _append(self, validated_func):
        """
        Adds the given `func` to the cycler to call.
        
        This method always runs on the cycler's event loop.
        
        Parameters
        ----------
        validated_func : ``CyclerCallable``
            The already validated function to add.
        """
        funcs = self.funcs
        funcs.append(validated_func)
        funcs.sort()
    
    
    def remove(self, func):
        """
        Removes the given `func` from the cycler, so it will stop calling it.
        
        If called from an other thread than it's event loop, then ensures ``._remove`` on it instead of calling it
        right away.
        
        Parameters
        ----------
        func : `callable`
            The function to remove.
        """
        loop = self.loop
        if current_thread() is loop:
            self._remove(func)
            return
        
        loop.call_soon_thread_safe_lazy(self.__class__._remove, self, func)
    
    
    def _remove(self, func):
        """
        Removes the given `func` from the cycler, so it will stop calling it.
        
        This method always runs on the cycler's event loop.
        
        Parameters
        ----------
        func : `callable`
            The function to remove.
        """
        index = 0
        funcs = self.funcs
        limit = len(funcs)
        
        is_cycler_callable = (type(func) is CyclerCallable)
        
        while index < limit:
            to_compare = funcs[index]
            if (not is_cycler_callable):
                to_compare = to_compare.func
            
            if to_compare == func:
                del funcs[index]
                break
            
            index += 1
            continue
    
    
    def get_time_till_next_call(self):
        """
        Returns how much time is left till the next cycle call.
        
        Might return `-1.` if the cycler is closed, or `0.` if the calls are taking place right now.
        
        Returns
        -------
        time_till_next_call : `float`
        """
        handle = self.handle
        if handle is None:
            return -1.0 # wont be be called
        
        at = handle.when - LOOP_TIME()
        
        if at < 0.0:
            return 0.0 # right now
        
        return at
    
    
    def get_time_of_next_call(self):
        """
        Returns when the next cycle call will be.
        
        Might return `-1.` if the cycler is closed.
        
        Returns
        -------
        ime_of_next_call : `float`
        """
        handle = self.handle
        if handle is None:
            return -1. # wont be be called
        return handle.when



__all__ = ('shield',)

from ...utils import set_docs

from .future import FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, FUTURE_STATE_RETRIEVED, Future


class _FutureChainer:
    """
    Chains a future's result into an other one used as a callback of the source future.
    
    Attributes
    ----------
    target : ``Future``
        The target future to chain the result into if applicable.
    """
    __slots__ = ('target',)
    
    def __init__(self, target):
        """
        Creates a new ``_FutureChainer`` with the given target future.
        
        Parameters
        ----------
        target : ``Future``
            The target future to chain the result into if applicable.
        """
        self.target = target
    
    if __debug__:
        def __call__(self, future):
            # remove chain remover
            target = self.target
            callbacks = target._callbacks
            for index in range(len(callbacks)):
                callback = callbacks[index]
                if (type(callback) is _ChainRemover) and (callback.target is future):
                    del callbacks[index]
                    break
            
            # set result
            state = future._state
            if state == FUTURE_STATE_FINISHED:
                future._state = FUTURE_STATE_RETRIEVED
                if future._exception is None:
                    target.set_result(future._result)
                else:
                    target.set_exception(future._exception)
                return
            
            if state == FUTURE_STATE_RETRIEVED:
                if future._exception is None:
                    target.set_result(future._result)
                else:
                    target.set_exception(future._exception)
                return
            
            # if state == FUTURE_STATE_CANCELLED: normally, but the future can be cleared as well.
            target.cancel()
    
    else:
        def __call__(self, future):
            # remove chain remover
            target = self.target
            callbacks = target._callbacks
            for index in range(len(callbacks)):
                callback = callbacks[index]
                if type(callback) is _ChainRemover and callback.target is future:
                    del callbacks[index]
                    break
            
            # set result
            if future._state == FUTURE_STATE_FINISHED:
                exception = future._exception
                if exception is None:
                    target.set_result(future._result)
                else:
                    target.set_exception(exception)
                return
            
            # if state == FUTURE_STATE_CANCELLED: normally, but the future can be cleared as well.
            target.cancel()
    
    set_docs(__call__,
        """
        Chains the source future's result into the target one.
        
        Parameters
        ----------
        future : ``Future``
            The source future to chain it's result from.
        """
    )

class _ChainRemover:
    """
    Removes the ``_FutureChainer`` callback of the source future if the chained target future is marked as done before.
    
    Parameters
    -------
    target : ``Future``
        The source future.
    """
    __slots__ = ('target',)
    
    def __init__(self, target):
        self.target = target
    
    def __call__(self, future):
        # remove chainer
        callbacks = self.target._callbacks
        for index in range(len(callbacks)):
            callback = callbacks[index]
            if (type(callback) is _FutureChainer) and (callback.target is future):
                del callbacks[index]
                if __debug__:
                    # because this might be the only place, where we retrieve the result, we will just silence it.
                    future.__silence__()
                break


def shield(awaitable, loop):
    """
    Protects the given `awaitable` from being cancelled.
    
    Parameters
    ----------
    awaitable : `awaitable`
        The awaitable object to shield.
    loop : ``EventThread``
        The event loop to run the `awaitable` on.
    
    Returns
    -------
    un_protected : ``Future``
        If the given `awaitable` is a finished task, returns it, else returns a ``Future``, to what the original
        awaitable's result will be chained to.
    
    Raises
    -------
    TypeError
        The given `awaitable` is not `awaitable`.
    """
    protected = loop.ensure_future(awaitable)
    if protected._state != FUTURE_STATE_PENDING:
        return protected # already done, we can return
    
    un_protected = Future(loop)
    protected._callbacks.append(_FutureChainer(un_protected))
    un_protected._callbacks.append(_ChainRemover(protected))
    return un_protected



__all__ = ()

import sys, warnings

from ....utils.trace import CONSOLE_LINE_CACHE, is_syntax_error, right_strip_syntax_error_line


PYTHON_COMPILE_FLAG_DONT_IMPLY_DEDENT = 1 << 9
PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT = 1 << 13
PYTHON_COMPILE_FLAG_ALLOW_INCOMPLETE_INPUT = 1 << 14


if sys.version_info < (3, 8):
    PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT = 0


try:
    compile('', '', 'single', PYTHON_COMPILE_FLAG_ALLOW_INCOMPLETE_INPUT)
except ValueError:
    PYTHON_COMPILE_FLAG_ALLOW_INCOMPLETE_INPUT = 0
except SyntaxError:
    pass


PYTHON_COMPILE_FLAGS_USED = (
    PYTHON_COMPILE_FLAG_DONT_IMPLY_DEDENT |
    PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT |
    PYTHON_COMPILE_FLAG_ALLOW_INCOMPLETE_INPUT
)


def maybe_compile(buffer, file_name):
    """
    Tries to compile the given source code.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The source code buffer to try to compile.
    file_name : `str`
        The file to compile the code under.
    
    Returns
    -------
    code : `None`, `CodeType`
    
    Raises
    ------
    SyntaxError
    """
    if is_source_empty(buffer):
        return None
    
    source = '\n'.join([line.rstrip() for line in buffer])
    
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        
        try:
            try:
                code_object = _try_compile(source, file_name, True)
            except IndentationError:
                raise
            
            except SyntaxError as err:
                exception_1 = err
            
            else:
                if (code_object is not None):
                    CONSOLE_LINE_CACHE.feed(file_name, source)
                    return code_object
                
                exception_1 = None
            
            try:
                code_object = _try_compile(source, file_name, False)
            except IndentationError:
                raise
            
            except SyntaxError as err:
                exception_2 = err
            
            else:
                if (code_object is not None):
                    CONSOLE_LINE_CACHE.feed(file_name, source)
                    return code_object
                
                exception_2 = None
            
            if (exception_1 is None):
                if (exception_2 is not None):
                    raise exception_2
            
            else:
                if (exception_2 is not None):
                    if is_syntax_error(exception_1):
                        right_strip_syntax_error_line(exception_1)
                    
                    if is_syntax_error(exception_2):
                        right_strip_syntax_error_line(exception_2)
                    
                    if exception_1.args == exception_2.args:
                        raise SyntaxError(*exception_1.args)
        
        finally:
            exception_1 = None
            exception_2 = None
    
    return None


def _try_compile(source, file_name, single_expression_mode):
    """
    Tries to compile the given source code.
    
    Parameters
    ----------
    source : `str`
        The source code to try to compile.
    file_name : `str`
        The file to compile the code under.
    single_expression_mode : `bool`
        Whether we expect a single expression.
    
    Returns
    -------
    code : `CodeType`
    
    Raises
    ------
    SyntaxError
    """
    try:
        return compile(
            source,
            file_name,
            'single' if single_expression_mode else 'exec',
            PYTHON_COMPILE_FLAGS_USED,
        )
    except SyntaxError as err:
        exception_representation = repr(err)
        
        if single_expression_mode:
            if 'multiple statements' in exception_representation:
                return None
        
        if PYTHON_COMPILE_FLAG_ALLOW_INCOMPLETE_INPUT:
            reason = 'incomplete input'
        else:
            reason = 'unexpected EOF'
        
        if reason in exception_representation:
            return None
        
        raise
    
    # Memory error if: too many nested parentheses
    # Value error if: cannot convert a value
    except (MemoryError, ValueError, OverflowError) as err:
        raise SyntaxError(*err.args) from err
    

def is_source_empty(buffer):
    """
    Returns whether the given source code is empty or nah.
    
    Parameters
    ----------
    buffer : `str`
        The source code buffer.
    
    Returns
    -------
    is_source_empty : `bool`
    """
    for line in buffer:
        for character in line:
            if character in {'\r', '\n', ' ', '\t'}:
                continue
        
            if character == '#':
                break
            
            return False
    
    return True



__all__ = ('modulize',)

import sys
from types import FunctionType, GetSetDescriptorType, MappingProxyType, ModuleType

from .docs import has_docs


NoneType = type(None)

try:
    from _weakref import ref as WeakrefType
except ImportError:
    from weakref import ref as WeakrefType

# This 2 type can be function
WrapperDescriptorType = type(object.__ne__)
MethodDescriptorType = type(object.__format__)

DO_NOT_MODULIZE_TYPES = [MappingProxyType, GetSetDescriptorType]

if WrapperDescriptorType is not FunctionType:
    DO_NOT_MODULIZE_TYPES.append(WrapperDescriptorType)

if MethodDescriptorType is not FunctionType:
    DO_NOT_MODULIZE_TYPES.append(MethodDescriptorType)

DO_NOT_MODULIZE_TYPES = tuple(DO_NOT_MODULIZE_TYPES)

@has_docs
def _modulize_function(old, globals_, source_module, module_name, module_path):
    """
    Changes the given function's scopes and qualname if they were defined inside of a modulized class.
    
    Parameters
    ----------
    old : `function`
        A function present inside of a modulized class.
    globals_ : `dict` of (`str`, `Any`)
        Global variables of the respective module.
    source_module : `module`
        The module, where the modulized class was defined.
    module_name : `str`
        The newly created module's name.
    module_path : `str`
        The newly created module's path.

    Returns
    -------
    new : `function`
        Newly recreated function if applicable.
    """
    if old.__module__ != source_module:
        return old
    
    new = FunctionType(old.__code__, globals_, old.__name__, old.__defaults__, old.__closure__)
    new.__module__ = module_path
    qualname = old.__qualname__
    if (qualname is not None) and (len(qualname) > len(module_name)) and qualname[len(module_name)] =='.' and \
            qualname.startswith(module_name):
        new.__qualname__ = qualname[len(module_name) + 1:]
    
    return new

@has_docs
def _modulize_type(klass, globals_, source_module, module_name, module_path):
    """
    Changes the given class's scopes and qualname if they were defined inside of a modulized class.
    
    Parameters
    ----------
    klass : `type`
        A class present inside of a modulized class.
    globals_ : `dict` of (`str`, `Any`)
        Global variables of the respective module.
    source_module : `module`
        The module, where the modulized class was defined.
    module_name : `str`
        The newly created module's name.
    module_path : `str`
        The newly created module's path.
    """
    if klass.__module__ != source_module:
        return
    
    qualname = klass.__qualname__
    if (qualname is None) or (len(qualname) <= len(module_name)) or qualname[len(module_name)] != '.' \
            or not qualname.startswith(module_name):
        return
    
    klass.__qualname__ = qualname[len(module_name) + 1:]
    klass.__module__ = module_path
    
    for name in dir(klass):
        value = getattr(klass, name)
        
        value_type = value.__class__
        if value_type is FunctionType:
            value = _modulize_function(value, globals_, source_module, module_name, module_path)
            setattr(klass, name, value)
        
        if issubclass(value_type, type):
            _modulize_type(value, globals_, source_module, module_name, module_path)

@has_docs
def modulize(klass):
    """
    Transforms the given class to a module.
    
    Every functions and classes defined inside of given class, which are also present at transformation as well, will
    have their global scope modified.
    
    Parameters
    ----------
    klass : `type`
        The class to transform to module.
    
    Returns
    -------
    result_module : `module`
        The created module object.
    
    Raises
    ------
    TypeError
        If `klass` is not given as `type`.
    """
    if not isinstance(klass, type):
        raise TypeError(
            f'Only types can be modulized, got {klass.__class__.__name__}; {klass!r}.'
    )
    
    source_module = klass.__module__
    module_name = klass.__name__
    module_path = f'{klass.__module__}.{module_name}'
    try:
        result_module = sys.modules['module_path']
    except KeyError:
        result_module = ModuleType(module_path)
        sys.modules[module_path] = result_module
        globals_ = result_module.__dict__
        globals_['__builtins__'] = __builtins__
    else:
        globals_ = result_module.__dict__
        collected_names = []
        for name in globals_.keys():
            if name.startswith('__') and name.endswith('__'):
                continue
            
            collected_names.append(name)
        
        for name in collected_names:
            del globals_[name]
        
        globals_['__doc__'] = None
    
    for name in type.__dir__(klass):
        if name.startswith('__') and name.endswith('__') and name != '__doc__':
            continue
        
        value = type.__getattribute__(klass, name)
        value_type = type(value)
        if value_type in DO_NOT_MODULIZE_TYPES:
            continue
        
        if value_type is FunctionType:
            value = _modulize_function(value, globals_, source_module, module_name, module_path)
        
        if issubclass(value_type, type):
            _modulize_type(value, globals_, source_module, module_name, module_path)
        
        ModuleType.__setattr__(result_module, name, value)
    
    return result_module



__all__ = ('create_future', 'create_task', 'get_current_task', 'get_tasks')

from ..traps import Future, Task

from .event_loop import _get_event_loop_is_current_thread, get_event_loop


def create_future(loop=None):
    """
    Creates s future bound to the local event loop.
    
    Parameters
    ----------
    loop : `None`, ``EventThread`` = `None`, Optional
        The event loop to create bound future to.
    
    Returns
    -------
    task : ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    loop, is_current_thread = _get_event_loop_is_current_thread(loop)
    future = Future(loop)
    if (not is_current_thread):
        loop.wake_up()
    
    return future


def create_task(coroutine, loop=None):
    """
    Creates a task on the local event loop.
    
    Parameters
    ----------
    coroutine : `GeneratorType`, `CoroutineType`
        The coroutine to create task from.
    loop : `None`, ``EventThread`` = `None`, Optional
        The event loop to schedule the created task on.
    
    Returns
    -------
    task : ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    loop, is_current_thread = _get_event_loop_is_current_thread(loop)
    task = Task(coroutine, loop)
    if (not is_current_thread):
        loop.wake_up()
    
    return task



def get_current_task(loop=None):
    """
    Returns the currently executed task.
    
    Parameters
    ----------
    loop : `None`, ``EventThread`` = `None`, Optional
        The event loop to get the current task of.
    
    Returns
    -------
    task : `None`, ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    if loop is None:
        loop = get_event_loop()
    
    return loop.current_task


def get_tasks(loop=None):
    """
    Returns the pending tasks.
    
    Parameters
    ----------
    loop : `None`, ``EventThread`` = `None`, Optional
        The event loop to get the pending tasks
    
    Returns
    -------
    task : `None`, ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    if loop is None:
        loop = get_event_loop()
    
    return loop.get_tasks()



__all__ = ('BaseMethodDescriptor', 'BaseMethodType',)

from .docs import has_docs
from .method_like import MethodLike
from .properties import docs_property


@has_docs
class BaseMethodType(MethodLike):
    """
    A `method-like`, which always passes to it's function the respective type and an instance. The instance might be
    given as `None` if used as a classmethod.
    
    Attributes
    ----------
    __base__ : `Any`
        The instance from where the method was called from. Might be `None` if used as a classmethod.
    __func__ : `function`
        The method's function to call.
    __self__ : `type`
        The class from where the method was called from.
    
    Class Attributes
    ----------------
    __reserved_argcount__ : `int` = `2`
        The amount of reserved parameters by BaseMethodTypes.
    """
    __slots__ = ('__base__', '__func__', '__self__', )
    
    __reserved_argcount__ = 2
    
    @has_docs
    def __init__(self, func, cls, base):
        """
        Creates a new BaseMethodType with the given parameters.
        
        Parameters
        ----------
        func : `function`
            The method's function to call.
        cls : `type`
            The class from where the method was called from.
        base : `Any`
            The instance from where the method was called from. Can be given as `None` as well.
        """
        self.__base__ = base
        self.__func__ = func
        self.__self__ = cls
    
    @has_docs
    def __call__(self, *args, **kwargs):
        """
        Calls the BaseMethodType with the given parameters.
        
        Parameters
        ----------
        *args : Parameters
            Parameters to call the internal function with.
        **kwargs : Keyword parameters
            Keyword parameters to call the internal function with.
        
        Returns
        -------
        result : `Any`
            The object returned by the internal function.
        
        Raises
        ------
        BaseException
            Exception raised by the internal function.
        """
        return self.__func__(self.__self__, self.__base__, *args, **kwargs)
    
    def __getattr__(self,name):
        return getattr(self.__func__, name)
    
    __class_doc__ = None
    
    @property
    @has_docs
    def __instance__doc__(self):
        """
        Returns the ``BaseMethodType``'s internal function's docstring.
        
        Returns
        -------
        docstring : `None`, `str`
        """
        return self.__func__.__doc__
    
    __doc__ = docs_property()

@has_docs
class BaseMethodDescriptor:
    """
    Descriptor, which can be used as a decorator to wrap a function to a BaseMethodType.
    
    Attributes
    ----------
    fget : `function`
        The wrapped function.
    """
    __slots__ = ('fget',)
    
    @has_docs
    def __init__(self, fget):
        """
        Creates a new ``BaseMethodDescriptor`` with the given parameter.
        
        Parameters
        ----------
        fget : `function`
            The function to wrap.
        """
        self.fget = fget
    
    def __get__(self, obj, type_):
        return BaseMethodType(self.fget, type_, obj)
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')



__all__ = ()

import socket as module_socket
import sys
from threading import current_thread

from ...utils import DOCS_ENABLED, docs_property, include

from ..traps import Task


EventThread = include('EventThread')

_HAS_IPv6 = hasattr(module_socket, 'AF_INET6')

def _ip_address_info(host, port, family, type_, protocol):
    """
    Gets the address info for the given parameters.
    
    Parameters
    ----------
    host : `str`, `bytes`
        The host ip address.
    port : `int`
        The host port.
    family : `AddressFamily`, `int`
        Address family.
    type_ : `SocketKind`, `int`
        Socket type.
    protocol : `int`
        Protocol type.
    
    Returns
    -------
    result : `None`, `tuple` ((`AddressFamily`, `int`), (`SocketKind`, `int`), `int`, `str`, `tuple` (`str, `int`))
        If everything is correct, returns a `tuple` of 5 elements:
        - `family` : Address family.
        - `type_` : Socket type.
        - `protocol` : Protocol type.
        - `canonical_name` : Represents the canonical name of the host. (Always empty string.)
        - `socket_address` : Socket address containing the `host` and the `port`.
    """
    # Try to skip get_address_info if `host` is already an IP. Users might have handled name resolution in their own
    # code and pass in resolved IP-s.
    if not hasattr(module_socket, 'inet_pton'):
        return
    
    if protocol not in (0, module_socket.IPPROTO_TCP, module_socket.IPPROTO_UDP) or (host is None):
        return
    
    if type_ == module_socket.SOCK_STREAM:
        protocol = module_socket.IPPROTO_TCP
    elif type_ == module_socket.SOCK_DGRAM:
        protocol = module_socket.IPPROTO_UDP
    else:
        return
    
    if port is None:
        port = 0
    elif type(port) is int:
        # Has the most chance.
        pass
    elif isinstance(port, bytes) and port == b'':
        port = 0
    elif isinstance(port, str) and port == '':
        port = 0
    else:
        # If port's a service name like "http", don't skip get_address_info.
        try:
            port = int(port)
        except (TypeError, ValueError):
            return
    
    if family == module_socket.AF_UNSPEC:
        address_families = [module_socket.AF_INET]
        if _HAS_IPv6:
            address_families.append(module_socket.AF_INET6)
    else:
        address_families = [family]
    
    if isinstance(host, bytes):
        host = host.decode('idna')
    
    if '%' in host:
        return
    
    for family in address_families:
        try:
            module_socket.inet_pton(family, host)
            # The host has already been resolved.
        except OSError:
            pass
        else:
            return family, type_, protocol,'', (host, port)
    
    # `host` is not an IP address.
    return None


def _is_datagram_socket(socket):
    """
    Returns whether the given socket is datagram socket.
    
    Parameters
    ----------
    socket : `socket.socket`
        The socket to check.
    
    Returns
    -------
    is_datagram_socket : `bool`
    """
    return (socket.type & module_socket.SOCK_DGRAM) == module_socket.SOCK_DGRAM


def _is_stream_socket(socket):
    """
    Returns whether the given socket is stream socket.
    
    Parameters
    ----------
    socket : `socket.socket`
        The socket to check.
    
    Returns
    -------
    is_stream_socket : `bool`
    """
    return (socket.type & module_socket.SOCK_STREAM) == module_socket.SOCK_STREAM

def _set_reuse_port(socket):
    """
    Tells to the kernel to allow this endpoint to be bound to the same port as an other existing endpoint already
    might be bound to.
    
    Parameters
    ----------
    socket : `socket.socket`.
        The socket to set reuse to.
    
    Raises
    ------
    ValueError
        `reuse_port` is not supported by socket module.
    """
    if not hasattr(module_socket, 'SO_REUSEPORT'):
        raise ValueError(
            '`reuse_port` not supported by socket module.'
        )
    else:
        try:
            socket.setsockopt(module_socket.SOL_SOCKET, module_socket.SO_REUSEPORT, 1)
        except OSError:
            raise ValueError(
                '`reuse_port` not supported by socket module, `SO_REUSEPORT` defined but not implemented.'
            )

_OLD_ASYNC_GENERATOR_HOOKS = sys.get_asyncgen_hooks()

def _async_generator_first_iteration_hook(async_generator):
    """
    Adds asynchronous generators to their respective event loop. These async gens are shut down, when the loop is
    stopped.
    
    Parameters
    ----------
    async_generator : `async_generator`
    """
    loop = current_thread()
    if isinstance(loop, EventThread):
        if loop._async_generators_shutdown_called:
            return
        
        loop._async_generators.add(async_generator)
        return
    
    first_iteration = _OLD_ASYNC_GENERATOR_HOOKS.firstiter
    if first_iteration is not None:
        first_iteration(async_generator)


def _async_generator_finalizer_hook(async_generator):
    """
    Removes asynchronous generator from their respective event loop.
    
    Parameters
    ----------
    async_generator : `async_generator`
    """
    loop = current_thread()
    if isinstance(loop, EventThread):
        loop._async_generators.discard(async_generator)
        if not loop.running:
            return
            
        Task(async_generator.aclose(), loop)
        loop.wake_up()
        return
    
    finalizer = _OLD_ASYNC_GENERATOR_HOOKS.finalizer
    if finalizer is not None:
        finalizer(async_generator)

sys.set_asyncgen_hooks(firstiter=_async_generator_first_iteration_hook, finalizer=_async_generator_finalizer_hook)


class EventThreadRunDescriptor:
    if DOCS_ENABLED:
        __class_doc__ = ("""
        Descriptor which decides, exactly which function of the ``EventThread`` is called, when using it's `.run`
        method.
        
        If called from class, returns `self`. If called from a non yet running event loop, returns that's `.runner`. If
        called from an already stopped event loop, raises `RuntimeError`.
        """)
        
        __instance_doc__ = ("""
        ``EventThread.run`` is an overloaded method, with two usages. The first is when the thread starts up, it will
        run the thread's "runner", ``EvenThread.runner``. The other one usage is, when the event loop is running, then
        it returns it's "caller", ``EvenThread.caller``.
        
        If the event loop is already closed, raises ``RuntimeError``.
        """)
        
        __doc__ = docs_property()
    
    def __get__(self, obj, type_):
        if obj is None:
            return self
        
        if obj.running:
            return obj.caller
        
        if not obj.started:
            if obj._started.is_set():
                obj.started = True
                return obj.runner
            else:
                obj._do_start()
                return obj.caller
        
        if not obj._is_stopped:
            return obj.caller
        
        raise RuntimeError(f'The {obj.__class__.__name__} is already stopped.')
    
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')



__all__ = ('enter_executor',)

from threading import current_thread

from ...utils import ignore_frame, include

from ..exceptions import CancelledError

from .future import FUTURE_STATE_CANCELLED, FUTURE_STATE_PENDING, Future
from .task import Task


ignore_frame(__spec__.origin, 'get_result', 'raise exception',)
ignore_frame(__spec__.origin, '__iter__', 'yield self',)

EventThread = include('EventThread')

class enter_executor:
    """
    Async context manager for moving a ``Task``'s section's execution to an executor thread.
    
    Usage
    -----
    ```py
    # Do async code
    async with enter_executor():
        # Do blocking stuff
        # Also async operation are still possible, but not recommended, because they cause my thread switch.
    
    # Do async code
    ```
    
    Attributes
    ----------
    _enter_future : `None`, ``Future``
        The future, what blocks the task's the execution, meanwhile the thread switch is taking place.
    _exit_future : `None`, ``Future``
        The future, what blocks the task's execution, meanwhile the thread is switching back.
    _waited_future : `None`, ``Future``
        Blocking future used inside of the task, meanwhile it is in executor.
    _task : `None`, ``Task``
    """
    __slots__ = ('_enter_future', '_exit_future', '_waited_future', '_task')
    
    def __init__(self):
        self._enter_future = None
        self._task = None
        self._exit_future=None
        self._waited_future = None
    
    
    async def __aenter__(self):
        """
        Moves the current tasks' execution to an executor thread.
        
        This method is a coroutine.
        
        Raises
        ------
        RuntimeError
            - Called from outside of an ``EventThread``.
            - Called from outside of a ``Task``.
        """
        thread = current_thread()
        if not isinstance(thread, EventThread):
            raise RuntimeError(
                f'`{self.__class__.__name__}` entered outside of `{EventThread.__name__}`, at {thread!r}.'
            )
        
        task = thread.current_task
        if task is None:
            raise RuntimeError(
                f'`{self.__class__.__name__}` entered outside of a `{Task.__name__}`.'
            )
        
        self._task = task
        loop = task._loop
        future = Future(loop)
        self._enter_future = future
        loop.call_soon(self._enter_executor)
        await future
        return self
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Moves the current task's executor back from an executor thread.
        
        This method is a coroutine.
        """
        await self._exit_future
        self._enter_future = None
        self._task = None
        self._exit_future = None
        self._waited_future = None
        return False
    
    
    def _enter_executor(self):
        """
        Moves the task's execution to an executor thread and wakes it up.
        """
        callbacks = self._enter_future._callbacks
        callbacks.clear()
        
        task = self._task
        task.add_done_callback(self._cancel_callback)
        
        task._loop.run_in_executor(self._executor_task)
    
    
    def _cancel_callback(self, future):
        """
        Callback added to the wrapped task. If the wrapped task is cancelled, then the section running inside of the
        executor will be cancelled as well, whenever it gives back the context with an `await`.
        """
        if future._state != FUTURE_STATE_CANCELLED:
            return
        
        waited_future = self._waited_future
        if waited_future is None:
            return
        
        waited_future.cancel()
    
    
    def _executor_task(self):
        """
        Wraps the tasks' section's running inside of an executor, still allowing it to use `await`-s.
        """
        task = self._task
        # relink future task
        loop = task._loop
        end_future = Future(loop)
        task._waited_future = end_future
        self._exit_future = end_future
        
        # Set result to the enter task, so it can be retrieved.
        self._enter_future.set_result(None)
        
        exception = None
        coroutine = task._coroutine
        
        # If some1 await at the block, we will sync_wrap it. If the exit future is awaited, then we quit.
        local_waited_future = None
        
        try:
            while True:
                if task._must_cancel:
                    exception = task._must_exception(exception)
                    
                if (local_waited_future is not None):
                    if local_waited_future is end_future:
                        end_future.set_result(None)
                        loop.call_soon_thread_safe(task._step, exception)
                        break
                    
                    try:
                        self._waited_future = local_waited_future
                        if type(exception) is CancelledError:
                            local_waited_future.cancel()
                        local_waited_future.sync_wrap().wait()
                    
                    except CancelledError:
                        break
                    except BaseException as err:
                        exception = err
                    finally:
                        local_waited_future = None
                        self._waited_future = None
                
                if task._state != FUTURE_STATE_PENDING:
                    # there is no reason to raise
                    break
                
                # call either coroutine.throw(err) or coroutine.send(None).
                try:
                    if exception is None:
                        result = coroutine.send(None)
                    else:
                        result = coroutine.throw(exception)
                
                except StopIteration as exception:
                    if task._must_cancel:
                        # the task is cancelled meanwhile
                        task._must_cancel = False
                        Future.set_exception(task, CancelledError())
                    else:
                        Future.set_result(task, exception.value)
                    
                    loop.wake_up()
                    break
                    
                except CancelledError:
                    Future.cancel(task)
                    loop.wake_up()
                    break
                
                except BaseException as exception:
                    Future.set_exception(task, exception)
                    loop.wake_up()
                    break
                
                else:
                    if isinstance(result, Future) and result._blocking:
                        result._blocking = False
                        local_waited_future = result
                        if task._must_cancel:
                            if local_waited_future.cancel():
                                task._must_cancel = False
                    else:
                        continue
        
        finally:
            task.remove_done_callback(self._cancel_callback)
            self = None
            task = None



__all__ = ('WebSocketCommonProtocol', )

import codecs, reprlib
import http as module_http
from collections import OrderedDict
from os import urandom

from ..core import AsyncQueue, CancelledError, Future, Lock, Task, future_or_timeout, shield, write_exception_async
from ..utils import include
from ..web_common import ConnectionClosed, HttpReadWriteProtocol, PayloadError, WebSocketFrame, WebSocketProtocolError
from ..web_common.websocket_frame import (
    WEBSOCKET_DATA_OPERATIONS, WEBSOCKET_OPERATION_BINARY, WEBSOCKET_OPERATION_CLOSE, WEBSOCKET_OPERATION_CONTINUOUS,
    WEBSOCKET_OPERATION_PING, WEBSOCKET_OPERATION_PONG, WEBSOCKET_OPERATION_TEXT
)


FORBIDDEN = module_http.HTTPStatus.FORBIDDEN
UPGRADE_REQUIRED = module_http.HTTPStatus.UPGRADE_REQUIRED
BAD_REQUEST = module_http.HTTPStatus.BAD_REQUEST
INTERNAL_SERVER_ERROR = module_http.HTTPStatus.INTERNAL_SERVER_ERROR
SERVICE_UNAVAILABLE = module_http.HTTPStatus.SERVICE_UNAVAILABLE
SWITCHING_PROTOCOLS = module_http.HTTPStatus.SWITCHING_PROTOCOLS

WEBSOCKET_STATE_CONNECTING = 1
WEBSOCKET_STATE_OPEN = 2
WEBSOCKET_STATE_CLOSING = 3
WEBSOCKET_STATE_CLOSED = 4

WEBSOCKET_KEY = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

EXTERNAL_CLOSE_CODES = (1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011,)

HTTPClient = include('HTTPClient')

DECODER = codecs.getincrementaldecoder('utf-8')(errors='strict')


class WebSocketCommonProtocol(HttpReadWriteProtocol):
    """
    WebSocket protocol base which implements common functions between the client and the server side.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `Any`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    _drain_lock : ``Lock``
        Asynchronous lock to ensure, that only `1` frame is written in `1` time.
    close_code : `int`
        The websocket's close code if applicable. Defaults to `0`.
    close_connection_task : `None`, ``Task`` of ``.close_connection``
        A task, what is present meanwhile the websocket is closing to avoid race condition.
    close_timeout : `float`
        The maximal duration in seconds what is waited for response after close frame is sent. Defaults to `10.0`.
    close_reason : `None`, `str`
        The reason, why the websocket was closed. Set only after the websocket is closed. Close reason might not be
        received tho.
    connection_lost_waiter : ``Future``
        A future, what's result is set as `None`, when the connection is closed. Used to wait for close frames.
        
        ``shield`` it if using from outside.
    extensions : `None` or (`list` of `Any`)
        WebSocket extensions. Defaults to `None`, if there is not any.
    host : `str`
        The respective server's address to connect to.
    max_queue : `None`, `int`
        Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of it is removed.
         Defaults to `None`.
    max_size : `int`
        Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised. Defaults to `67108864` bytes.
    messages : ``AsyncQueue``
        An asynchronous queue of the received messages.
    is_ssl : `bool`
        Whether the connection is secure. Defaults to `False`.
    pings : `OrderedDict` of (`bytes`, ``Future``) items
        An ordered dictionary of ping payloads and of their waiter futures.
    port : `int`
        The respective server's port to connect to.
    state : `str`
        The websocket's state.
        
        Can be set as one of the following values:
        
        +-------------------------------+-------------------+
        | Respective name               | Value             |
        +===============================+===================+
        | WEBSOCKET_STATE_CONNECTING    | `1`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_OPEN          | `2`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_CLOSING       | `3`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_CLOSED        | `4`               |
        +-------------------------------+-------------------+
        
        Note, that state == compared by memory address and not by value.
    subprotocol : `None`, `str`
        Chosen subprotocol at handshake. Defaults to `None` and might be set as `str`. Chosen from the available
        subprotocols by their priority order.
    transfer_data_exception : `None`, `BaseException``
        Exception catched meanwhile processing received data.
    transfer_data_task : `None`, ``Task`` of ``.transfer_data``
        Data receiving task.
    
    Class Attributes
    ----------------
    is_client : `bool` = `True`
        Whether the websocket protocol is client or server side.
    """
    __slots__ = (
        '_drain_lock', 'close_code', 'close_connection_task', 'close_timeout', 'close_reason',
        'connection_lost_waiter', 'extensions', 'host', 'is_ssl', 'max_queue', 'max_size', 'messages', 'pings', 'port',
        'state', 'subprotocol', 'transfer_data_exception', 'transfer_data_task'
    )
    
    is_client = True # placeholder for subclasses
    
    def __new__(cls, loop, host, port, *, is_ssl=False, close_timeout=10.0, max_size=1 << 26, max_queue=None):
        """
        Initializes the ``WebSocketCommonProtocol`` with setting it's common attributes.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop, what the protocol uses for it's asynchronous tasks.
        host : `str`
            The respective server's address to connect to.
        port : `int`
            The respective server's port to connect to.
        is_ssl : `bool` = `False`, Optional (Keyword only)
            Whether the connection is secure.
        close_timeout : `float` = `10.0`, Optional (Keyword only)
            The maximal duration in seconds what is waited for response after close frame is sent
        max_size : `int` = `67108864`, Optional (Keyword only)
            Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised.
        max_queue : `None`, `int` = `None`, Optional (Keyword only)
            Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of it is
            removed.
        """
        self = HttpReadWriteProtocol.__new__(cls, loop)
        self._set_common_websocket_attributes(host, port, is_ssl, close_timeout, max_size, max_queue)
        return self
    
    def _set_common_websocket_attributes(self, host, port, is_ssl, close_timeout=10.0, max_size=1 << 26, max_queue=None):
        """
        Sets the common websocket specific attributes for the protocol.
        
        Parameters
        ----------
        host : `str`
            The respective server's address to connect to.
        port : `int`
            The respective server's port to connect to.
        is_ssl : `bool`
            Whether the connection is secure. Defaults to `False`.
        close_timeout : `float` = `10.0`, Optional
            The maximal duration in seconds what is waited for response after close frame is sent.
        max_size : `int` = `67108864`, Optional
            Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised.
            bytes.
        max_queue : `None`, `int` = `None`, Optional
            Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of it is
            removed.
        """
        self.host = host
        self.port = port
        self.is_ssl = is_ssl
        self.close_timeout = close_timeout
        self.max_size = max_size # set it to a BIG number if u wanna ignore max size
        self.max_queue = max_queue
        
        self._drain_lock = Lock(self._loop)
        
        self.state = WEBSOCKET_STATE_CONNECTING
        
        self.extensions = None # set from outside
        self.subprotocol = None # set from outside
        
        self.close_code = 0
        self.close_reason = None
        
        self.connection_lost_waiter = Future(self._loop)
        self.messages = AsyncQueue(loop=self._loop, max_length=max_queue)
        
        self.pings = OrderedDict()
        
        self.transfer_data_task = None
        self.transfer_data_exception = None
        self.close_connection_task = None
    
    
    def connection_open(self):
        """
        Method called when the connection is established at the end of the handshake.
        
        Marks the websocket as open and start it's ``.transfer_data_task`` and ``.close_connection_task``.
        """
        self.state = WEBSOCKET_STATE_OPEN
        loop = self._loop
        self.transfer_data_task = Task(self.transfer_data(), loop)
        self.close_connection_task = Task(self.close_connection(), loop)
    
    
    @property
    def local_address(self):
        """
        Local address of the connection as a `tuple` of host and port. If the connection is not open yet, returns
        `None`.
        
        Returns
        -------
        local_address : `None`, `tuple` of (`str`, `int`)
        """
        return self.get_extra_info('socket_name')
    
    
    @property
    def remote_address(self):
        """
        Remote address of the connection as a `tuple` of host and port. If the connection is not open yet, returns
        `None`.
        
        Returns
        -------
        remote_address : `None`, `tuple` of (`str`, `int`)
        """
        return self.get_extra_info('peer_name')
    
    
    @property
    def open(self):
        """
        Returns whether the websocket is open.
        
        If the websocket is closed, ``ConnectionClosed`` is raised when using it.
        
        Returns
        -------
        open : `bool`
        """
        if self.state != WEBSOCKET_STATE_OPEN:
            return False
        
        transfer_data_task = self.transfer_data_task
        if transfer_data_task is None:
            return False
        
        if self.transfer_data_task.is_done():
            return False
        
        return True
    
    
    @property
    def closed(self):
        """
        Returns whether the websocket is closed.
        
        Note, meanwhile connection is establishing, ``.open`` and ``.close`` will return `False`.
        
        Returns
        -------
        closed : `bool`
        """
        return self.state == WEBSOCKET_STATE_CLOSED
    
    
    def receive(self):
        """
        Returns a future, what can be awaited to receive the next message of the websocket.
        
        Returns
        -------
        future : ``Future``
            The future returns `bytes`, `str` respective to the received payload's type. If the websocket
            is closed, ``ConnectionClosed`` is raised.
        """
        return self.messages.get_result()
    
    def receive_no_wait(self):
        """
        Returns a future, what can be awaited to receive the next message of the websocket.
        
        Returns
        -------
        message : `bytes`, `str`
            The received payload.
        
        Raises
        ------
        IndexError
            There are no messages to retrieve right now.
        ConnectionClosed
            WebSocket closed.
        """
        return self.messages.result_no_wait()
    
    async def send(self, data):
        """
        Sends the given data with the websocket.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `bytes-like`, `str`
            The data to send
        
        Raises
        ------
        TypeError
            `data` was not given as `bytes-like`, neither as `str`.
        ConnectionClosed
            WebSocket connection closed.
        Exception
            WebSocket connection not yet established.
        """
        await self.ensure_open()
        
        if isinstance(data, (bytes, bytearray, memoryview)):
            operation_code = WEBSOCKET_OPERATION_BINARY
        elif isinstance(data, str):
            operation_code = WEBSOCKET_OPERATION_TEXT
            data = data.encode('utf-8')
        else:
            raise TypeError(
                f'Data must be `bytes-like`, `str`, got: {data.__class__.__name__}; {reprlib.repr(data)}.'
            )

        await self.write_frame(operation_code, data)
    
    
    async def close(self, code=1000, reason=''):
        """
        Closes the websocket.
        
        Writes close frame first and then if we don't receive close frame response in ``.close_timeout``, then we
        cancel the connection.
        
        This method is a coroutine.
        
        Parameters
        ----------
        code : `int` = `1000`, Optional.
            WebSocket close code. Defaults to `1000`. Can be one of:
            `(1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011) | [3000:5000)`.
        reason : `str` = `''`, Optional
            WebSocket close reason. Can be given as empty string.
        
        Raises
        ------
        WebSocketProtocolError
            Close code is not given as one of `(1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011) | [3000:5000)`.
        """
        # if no close frame is received within the close_timeout we cancel the connection
        close_message = self._serialize_close(code, reason)
        try:
            task = Task(self.write_close_frame(close_message), self._loop)
            future_or_timeout(task, self.close_timeout)
            await task
        except TimeoutError:
            self.fail_connection()
        
        try:
            # if close() is cancelled during the wait, self.transfer_data_task is cancelled before the close_timeout
            # elapses
            task = self.transfer_data_task
            future_or_timeout(task, self.close_timeout)
            await task
        except (TimeoutError, CancelledError):
            pass
        
        # quit for the close connection task to close the TCP connection.
        await shield(self.close_connection_task, self._loop)
    
    @staticmethod
    def _serialize_close(code, reason):
        """
        Packs the given `code` and `reason` together into a close message.
        
        Parameters
        ----------
        code : `int`
            WebSocket close code. Can be one of:
            `(1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011) | [3000:5000)`.
        reason : `str`
            WebSocket close reason. Can be given as empty string.
        
        Returns
        -------
        close_message : `bytes`
        
        Raises
        ------
        WebSocketProtocolError
            Close code is not given as one of `(1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011) | [3000:5000)`.
        """
        if not (code in EXTERNAL_CLOSE_CODES or 2999<code<5000):
            raise WebSocketProtocolError(
                f'Status can be in range [3000:5000), got {code!r}.'
            )
        
        return code.to_bytes(2, 'big') + reason.encode('utf-8')
    
    async def ping(self, data=None):
        """
        Sends a ping to the other side and waits till answer is received.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `None`, `bytes-like`, `str`
            Ping payload to send. Defaults to `None`.
            
            If the given or generated payload is already waiting for response, then will regenerate it, till a free
            one is hit.
        
        Raises
        ------
        TypeError
            `data` is not given neither as `None`, `bytes-like`, `str`.
        ConnectionClosed
            WebSocket connection closed.
        Exception
            WebSocket connection not yet established.
        """
        await self.ensure_open()
        
        if data is None:
            data = urandom(4)
        elif isinstance(data, (bytes, bytearray, memoryview)):
            pass
        elif isinstance(data, str):
            data = data.encode('utf-8')
        else:
            raise TypeError(
                f'Data must be `bytes-like`, `str`, got: {data.__class__.__name__}; {reprlib.repr(data)}.'
            )
        
        pings = self.pings
        while data in pings:
            data = urandom(4)
        
        waiter = Future(self._loop)
        pings[data] = waiter
        
        await self.write_frame(WEBSOCKET_OPERATION_PING, data)
        
        await waiter
    
    async def pong(self, data=None):
        """
        Sends a pong payload to the other side.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `None`, `bytes-like`, `str`
            Ping payload to send. Defaults to `None`.
        
        Raises
        ------
        TypeError
            `data` is not given neither as `None`, `bytes-like`, `str`.
        ConnectionClosed
            WebSocket connection closed.
        Exception
            WebSocket connection not yet established.
        """
        await self.ensure_open()
        
        if data is None:
            data = urandom(4)
        elif isinstance(data, (bytes, bytearray, memoryview)):
            pass
        elif isinstance(data, str):
            data = data.encode('utf-8')
        else:
            raise TypeError(
                f'Data must be `bytes-like`, `str`, got: {data.__class__.__name__}; {reprlib.repr(data)}.'
            )
        
        await self.write_frame(WEBSOCKET_OPERATION_PONG, data)
    
    # Private methods - no guarantees.
    
    async def ensure_open(self):
        """
        Checks whether the websocket is still open.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionClosed
            WebSocket connection closed.
        Exception
            WebSocket connection not yet established.
        """
        state = self.state
        if state == WEBSOCKET_STATE_OPEN:
            # if self.transfer_data_task exited without a closing handshake.
            if self.transfer_data_task.is_done():
                await shield(self.close_connection_task, self._loop)
                raise ConnectionClosed(self.close_code, None, self.close_reason)
            
            return
        
        if state == WEBSOCKET_STATE_CLOSED:
            raise ConnectionClosed(self.close_code, None, self.close_reason) from self.transfer_data_exception
        
        if state == WEBSOCKET_STATE_CLOSING:
            if not self.close_code:
                #if we started the closing handshake, wait for its completion
                await shield(self.close_connection_task, self._loop)
            
            raise ConnectionClosed(self.close_code, None, self.close_reason) from self.transfer_data_exception
        
        raise Exception('WebSocket connection isn\'t established yet.')
    
    
    async def transfer_data(self):
        """
        The transfer data task of a websocket keeps reading it's messages and putting it into it's ``.messages``
        ``AsyncQueue``.
        
        Meanwhile runs, it wrapped inside of a ``Task`` and can be accessed as ``.transfer_data_task``.
        
        This method is a coroutine.
        """
        try:
            while True:
                message = await self.read_message()
                # exit the loop when receiving a close frame.
                if message is None:
                    break
                
                self.messages.set_result(message)
        
        except GeneratorExit as err:
            exception = ConnectionClosed(1013, err)
            self.fail_connection(self.close_code or 1013)
            to_reraise = err
        
        except CancelledError as err:
            # we already failed connection
            exception = ConnectionClosed(self.close_code or 1000, err, self.close_reason)
            to_reraise = err
        
        except WebSocketProtocolError as err:
            exception = ConnectionClosed(1002, err)
            self.fail_connection(1002)
            to_reraise = None
        
        except (ConnectionError, EOFError, TimeoutError) as err:
            exception = ConnectionClosed(1006, err)
            self.fail_connection(1006)
            to_reraise = None
        
        except UnicodeDecodeError as err:
            exception = ConnectionClosed(1007, err)
            self.fail_connection(1007)
            to_reraise = None
        
        except PayloadError as err:
            exception = ConnectionClosed(1009, err)
            self.fail_connection(1009)
            to_reraise = None
        
        except BaseException as err:
            await write_exception_async(
                err,
                [
                    'Unexpected exception occurred at ',
                    repr(self),
                    '.transfer_data\n',
                ],
                loop = self._loop,
            )
            # should not happen
            exception = ConnectionClosed(1011, err)
            self.fail_connection(1011)
            to_reraise = None
        
        else:
            # connection was closed
            exception = ConnectionClosed(self.close_code or 1000, None, self.close_reason)
            to_reraise = None
            
            # If we are a client and we receive this, we closed our own
            # connection, there is no reason to wait for TCP abort
            if self.is_client:
                self.connection_lost_waiter.set_result_if_pending(None)
        
        if self.transfer_data_exception is None:
            self.transfer_data_exception = exception
            self.messages.set_exception(exception)
        
        if (to_reraise is not None):
            raise to_reraise
    
    
    async def read_message(self):
        """
        Reads a message from the websocket.
        
        This method is a coroutine.
        
        Returns
        -------
        message : `None`, `bytes`, `str`
            A received message. It's type depend on the frame's type. returns `None` if close frame was received.
        
        Raises
        ------
        WebSocketProtocolError
            - Unexpected op code.
            - Incomplete fragmented message.
            - If the reserved bits are not `0`.
            - If the frame is a control frame, but is too long for one.
            - If the websocket frame is fragmented frame. (Might be supported if people request is.)
            - If the frame operation_code is not any of the expected ones.
            - Close frame received with invalid status code.
            - Close frame too short.
        CancelledError
            ``.transfer_data_task`` cancelled.
        """
        frame = await self.read_data_frame(max_size=self.max_size)
        if frame is None: # close frame
            return
        
        if frame.operation_code == WEBSOCKET_OPERATION_TEXT:
            text = True
        elif frame.operation_code == WEBSOCKET_OPERATION_BINARY:
            text = False
        else: # frame.operation_code == OP_CONT:
            raise WebSocketProtocolError(
                f'Unexpected operation_code, got {frame.operation_code!r}, expected {WEBSOCKET_OPERATION_TEXT!r} or '
                f'{WEBSOCKET_OPERATION_BINARY!r}.'
            )
        
        # we got a whole frame, nice
        if frame.is_final:
            message = frame.data
            
            if text:
                message = message.decode('utf-8')
            
            return message
        
        max_size = self.max_size # set max size to BIG number to ignore it
        
        frames = []
        while True:
            max_size -= len(frame.data)
            
            frames.append(frame)
            if frame.is_final:
                break
            
            frame = await self.read_data_frame(max_size=max_size)
            if frame is None:
                raise WebSocketProtocolError('Incomplete fragmented message.')
            
            if frame.operation_code != WEBSOCKET_OPERATION_CONTINUOUS:
                raise WebSocketProtocolError(
                    f'Unexpected operation_code, got {frame.operation_code!r}, expected '
                    f'{WEBSOCKET_OPERATION_CONTINUOUS!r}.'
                )
        
        if text:
            try:
                message = ''.join(DECODER.decode(frame.data, frame.is_final) for frame in frames)
            except:
                DECODER.reset()
                raise
        else:
            message = b''.join(frame.data for frame in frames)
        
        return message
    
    
    async def read_data_frame(self, max_size):
        """
        Reads a websocket frame from the websocket. If the frame is a control frame processes and loops for reading an
        another one.
        
        This method is a coroutine.
        
        Parameters
        ----------
        
        Returns
        -------
        frame : ``WebSocketFrame``, `None`
            The read websocket frame. Returns `None` if close frame was received.
        
        Raises
        ------
        WebSocketProtocolError
            - If the reserved bits are not `0`.
            - If the frame is a control frame, but is too long for one.
            - If the websocket frame is fragmented frame. (Might be supported if people request is.)
            - If the frame operation_code is not any of the expected ones.
            - Close frame received with invalid status code.
            - Close frame too short.
        CancelledError
            ``.transfer_data_task`` cancelled.
        """
        while True:
            
            frame = await self.set_payload_reader(self._read_websocket_frame(self.is_client, max_size))
            
            extensions = self.extensions
            if (extensions is not None):
                for extension in reversed(extensions):
                    frame = extension.decode(frame, max_size=max_size)
            
            frame.check()
            
            # most likely
            if frame.operation_code in WEBSOCKET_DATA_OPERATIONS:
                return frame

            if (await self._process_CONTROL_frame(frame)):
                continue
            
            return
    
    async def _process_CONTROL_frame(self, frame):
        """
        Processes a control websocket frame.
        
        This method is a coroutine.
        
        Parameters
        ----------
        frame : ``WebSocketFrame``
            A received control websocket frame.
        
        Returns
        -------
        can_continue : `bool`
            Returns `False` if the processed `frame` was a close frame.
        
        Raises
        ------
        WebSocketProtocolError
            - Close frame received with invalid status code.
            - Close frame too short.
        """
        operation_code = frame.operation_code
        if operation_code == WEBSOCKET_OPERATION_CLOSE:
            data = frame.data
            length = len(data)
            if length >= 2:
                code = int.from_bytes(data[:2], 'big')
                if not (code in EXTERNAL_CLOSE_CODES or 2999 < code < 5000):
                    raise WebSocketProtocolError(f'Invalid status code {code!r}.')
                reason = data[2:].decode('utf-8')
                self.close_code = code
                self.close_reason = reason
            elif length == 0:
                self.close_code = 1005
            else:
                raise WebSocketProtocolError(
                    f'Close frame too short: {length!r}; {reprlib.repr(data)}.'
                )
            
            await self.write_close_frame(frame.data)
            return False
        
        if operation_code == WEBSOCKET_OPERATION_PING:
            await self.pong(frame.data)
            return True
        
        # operation_code == OP_PONG:
        if frame.data in self.pings:
            #checking all pings up to the one matching this pong.
            ping_id = b''
            while ping_id != frame.data:
                ping_id, pong_waiter = self.pings.popitem(0)
                pong_waiter.set_result_if_pending(None)
        
        return True
    
    async def write_frame(self, operation_code, data, _expected_state=WEBSOCKET_STATE_OPEN):
        """
        Writes the data as websocket.
        
        This method is a coroutine.
        
        Parameters
        ----------
        operation_code : `int`
            The operation code of the websocket frame.
            
            Can be 1 of the following:
            
            +-----------------------------------+-------+
            | Respective name                   | Value |
            +===================================+=======+
            | WEBSOCKET_OPERATION_CONTINUOUS    | 0     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_TEXT          | 1     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_BINARY        | 2     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_CLOSE         | 8     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PING          | 9     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PONG          | 10    |
            +-----------------------------------+-------+
        
        data : `bytes-like`
            The data to send.
        
        _expected_state : `str`
            Expected state of the websocket. If the websocket is in other state, an `Exception` it raised.
            Defaults to `'WEBSOCKET_STATE_OPEN'`.
            
            Can be set as one of the following values:
            
            +-------------------------------+-------------------+
            | Respective name               | Value             |
            +===============================+===================+
            | WEBSOCKET_STATE_CONNECTING    | `1`               |
            +-------------------------------+-------------------+
            | WEBSOCKET_STATE_OPEN          | `2`               |
            +-------------------------------+-------------------+
            | WEBSOCKET_STATE_CLOSING       | `3`               |
            +-------------------------------+-------------------+
            | WEBSOCKET_STATE_CLOSED        | `4`               |
            +-------------------------------+-------------------+
            
            Note, that state == compared by memory address and not by value.
        
        Raises
        ------
        WebSocketProtocolError
            - If an extension set a reserved bit not to `0`.
            - If an extension modified the frame to a control frame, what is too long one.
            - If an extension modified the frame to be a fragmented one. (Might be supported if people request is.)
            - If an extension modified the frame's op code to not any of the expected ones.
        ConnectionClosed
            WebSocket connection closed.
        Exception
            - WebSocket connection not yet established.
            - Cannot write to websocket with it's current state.
        RuntimeError
            Protocol has no attached transport.
        """
        # Defensive assertion for protocol compliance.
        if self.state != _expected_state:
            raise Exception(f'Cannot write to a WebSocket in the {self.state} state.')

        # we write only 1 frame at a time, so we 'queue' it
        async with self._drain_lock:
            try:
                frame = WebSocketFrame(True, operation_code, data)
                
                extensions = self.extensions
                if (extensions is not None):
                    for extension in extensions:
                        frame = extension.encode(frame)
                    
                    frame.check()
                
                self.write_websocket_frame(frame, self.is_client)
                await self.drain()
            except ConnectionError:
                self.fail_connection()
                #raise ConnectionClosed with the correct code and reason.
                await self.ensure_open()
    
    
    async def write_close_frame(self, data=b''):
        """
        Writes close frame to the websocket if the websocket is not yet closed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `bytes-like` = `b''`, Optional
            The data to send.
        
        Raises
        ------
        WebSocketProtocolError
            - If an extension set a reserved bit not to `0`.
            - If an extension modified the frame to a control frame, what is too long one.
            - If an extension modified the frame to be a fragmented one. (Might be supported if people request is.)
            - If an extension modified the frame's op code to not any of the expected ones.
        ConnectionClosed
            WebSocket connection closed.
        Exception
            - WebSocket connection not yet established.
            - Cannot write to websocket with it's current state.
        RuntimeError
            Protocol has no attached transport.
        """
        # check connection before we write
        if self.state == WEBSOCKET_STATE_OPEN:
            self.state = WEBSOCKET_STATE_CLOSING
            await self.write_frame(WEBSOCKET_OPERATION_CLOSE, data, WEBSOCKET_STATE_CLOSING)
    
    
    async def close_connection(self):
        """
        Makes sure that the websocket is closed correctly.
        
        Meanwhile the websocket is closing, it's ``.close_connection_task`` is set as a ``Task`` object wrapping the
        ``.close_connection`` coroutine to avoid race condition.
        
        This method is a coroutine.
        """
        try:
            # Wait for the data transfer phase to complete.
            transfer_data_task = self.transfer_data_task
            if (transfer_data_task is not None):
                try:
                    await transfer_data_task
                except (CancelledError, TimeoutError):
                    pass
            
            # Cancel all pending pings because they'll never receive a pong.
            for ping in self.pings.values():
                ping.cancel()
            
            # A client should wait for a TCP close from the server.
            if self.is_client and (self.transfer_data_task is not None):
                if (await self.wait_for_connection_lost()):
                    return
            
            if self.can_write_eof():
                self.write_eof()
                if (await self.wait_for_connection_lost()):
                    return
        finally:
            # finally ensures that the transport never remains open
            if self.connection_lost_waiter.is_done() and not self.is_ssl:
                return
            
            # Close the TCP connection
            transport = self._transport
            if (transport is not None):
                transport.close()
                if (await self.wait_for_connection_lost()):
                    return
                # Abort the TCP connection
                transport.abort()
            
            
            # connection_lost() is called quickly after aborting.
            await self.wait_for_connection_lost()
    
    
    async def wait_for_connection_lost(self):
        """
        Waits until ``.connection_lost_waiter`` is set. If ``.close_timeout`` is over before it happens, returns.
        
        This method is a coroutine.
        
        Returns
        -------
        is_connection_lost : `bool`
            Returns `True` if the connection is lost.
        """
        if self.connection_lost_waiter.is_pending():
            try:
                task = shield(self.connection_lost_waiter, self._loop)
                future_or_timeout(task, self.close_timeout)
                await task
            except TimeoutError:
                pass
        
        # re-check self.connection_lost_waiter.is_done() synchronously because connection_lost() could run between the
        # moment the timeout occurs and the moment this coroutine resumes running.
        return self.connection_lost_waiter.is_done()
    
    def fail_connection(self, code=1006, reason=''):
        """
        Closes the websocket if any unexpected exception occurred.
        
        Parameters
        ----------
        code : `int` = `1006`, Optional
            WebSocket close code.
        reason : `str` = `''`, Optional
            WebSocket close reason.
        
        Returns
        -------
        close_connection_task : ``Task`` of ``.close_connection``
            Close connection task, what can be awaited to wait till the connection is closed.
        
        Raises
        ------
        WebSocketProtocolError
            - If an extension set a reserved bit not to `0`.
            - If an extension modified the frame to a control frame, what is too long one.
            - If an extension modified the frame to be a fragmented one. (Might be supported if people request is.)
            - If an extension modified the frame's op code to not any of the expected ones.
        RuntimeError
            Protocol has no attached transport.
        """
        # cancel transfer_data_task if the opening handshake succeeded
        transfer_data_task = self.transfer_data_task
        if transfer_data_task is not None:
            transfer_data_task.cancel()
        
        # send a close frame when the state == WEBSOCKET_STATE_OPEN and the connection is not broken
        if code != 1006 and self.state == WEBSOCKET_STATE_OPEN:
            
            frame_data = self._serialize_close(code, reason)
            # Write the close frame without draining the write buffer.
            self.state = WEBSOCKET_STATE_CLOSING
            
            frame = WebSocketFrame(True, WEBSOCKET_OPERATION_CLOSE, frame_data)
            
            extensions = self.extensions
            if (extensions is not None):
                for extension in extensions:
                    frame = extension.encode(frame)
                
                frame.check()
            
            self.write_websocket_frame(frame, self.is_client)
        
        # start close_connection_task if the opening handshake didn't succeed.
        close_connection_task = self.close_connection_task
        if close_connection_task is None:
            close_connection_task = Task(self.close_connection(), self._loop)
            self.close_connection_task = close_connection_task
        
        return close_connection_task
    
    # compatibility method (overwrite)
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self.state = WEBSOCKET_STATE_CLOSED
        if not self.close_code:
            self.close_code = 1006
        
        # `self.connection_lost_waiter` should be pending
        self.connection_lost_waiter.set_result_if_pending(None)
        
        HttpReadWriteProtocol.connection_lost(self, exception)
    
    # compatibility method (overwrite)
    def eof_received(self):
        """
        Calling ``.connection_lost`` without exception causes eof.
        
        Marks the protocols as it is at eof and stops payload processing if applicable.
        
        Returns
        -------
        transport_closes : `bool`
            Returns `False` if the transport will close itself. If it returns `True`, then closing the transport is up
            to the protocol.
            
            Returns `True` if the websocket is not secure.
        """
        HttpReadWriteProtocol.eof_received(self)
        return (not self.is_ssl)



__all__ = (
    'is_awaitable', 'is_coroutine', 'is_coroutine_function', 'is_coroutine_generator',
    'is_coroutine_generator_function', 'is_generator', 'is_generator_function', 'to_coroutine'
)

import sys
from types import AsyncGeneratorType, CoroutineType, FunctionType, GeneratorType, MethodType

from .code import CO_ASYNC_GENERATOR, CO_COROUTINE_ALL, CO_GENERATOR, CO_ITERABLE_COROUTINE
from .docs import has_docs, set_docs


FUNCTION_TYPES = (FunctionType, MethodType)


@has_docs
def is_coroutine_function(func):
    """
    Returns whether the given object is a coroutine function, so is created with `async def`.
    
    Parameters
    ----------
    func : `Any`
    
    Returns
    -------
    is_coroutine_function : `bool`
    """
    if isinstance(func, FUNCTION_TYPES) and func.__code__.co_flags & CO_COROUTINE_ALL:
        return True
    else:
        return False


@has_docs
def is_coroutine_generator_function(func):
    """
    Returns whether the given object is a coroutine generator function, so is created with `async def` and uses `yield`
    statement.
    
    Parameters
    ----------
    func : `Any`
    
    Returns
    -------
    is_coroutine_function_generator : `bool`
    """
    if isinstance(func, FUNCTION_TYPES) and func.__code__.co_flags & CO_ASYNC_GENERATOR:
        return True
    else:
        return False


@has_docs
def is_generator_function(func):
    """
    Returns whether the given `func` is a generator function, so is created with `def` and uses `yield` statement.
    
    Parameters
    ----------
    func : `Any`
    
    Returns
    -------
    is_generator_function : `bool`
    """
    if isinstance(func, FUNCTION_TYPES) and func.__code__.co_flags & CO_GENERATOR:
        return True
    else:
        return False


@has_docs
def is_coroutine(obj):
    """
    Returns whether the given object is a coroutine, so is created with an `async def`.
    
    Parameters
    ----------
    obj : `Any`
    
    Returns
    -------
    is_coroutine : `bool`
    """
    if isinstance(obj, CoroutineType):
        return True

    if isinstance(obj, GeneratorType):
        if obj.gi_code.co_flags & CO_ITERABLE_COROUTINE:
            return True
        else:
            return False
    
    return False


@has_docs
def is_awaitable(obj):
    """
    Returns whether the given object can be used in `await` expression.
    
    Parameters
    ----------
    obj : `Any`
    
    Returns
    -------
    is_awaitable : `bool`
    """
    if is_coroutine(obj):
        return True
    
    if hasattr(type(obj), '__await__'):
        return True
    
    return False


@has_docs
def is_coroutine_generator(obj):
    """
    Returns whether the given object is a coroutine generator, so is created with `async def` and uses `yield`
    statement, and could be used inside of an `async for` loop.
    
    Parameters
    ----------
    obj : `Any`
    
    Returns
    -------
    is_coroutine_generator : `bool`
    """
    if isinstance(obj, AsyncGeneratorType):
        code = obj.ag_code
    elif isinstance(obj, CoroutineType):
        code = obj.cr_code
    elif isinstance(obj, GeneratorType):
        code = obj.gi_code
    else:
        return False
    
    if code.co_flags & CO_ASYNC_GENERATOR:
        return True
    
    return False


@has_docs
def is_generator(obj):
    """
    Returns whether the given object is a generator, so is created with an `def` function and uses `yield` statement,
    and could be used inside of a `for` loop.
    
    Parameters
    ----------
    obj : `Any`
    
    Returns
    -------
    is_coroutine_generator : `bool`
    """
    return isinstance(obj, GeneratorType)


if sys.version_info >= (3, 11, 0):
    def to_coroutine(function):
        if not isinstance(function, FunctionType):
            raise TypeError(
                f'`function` can only be `{FunctionType.__name__}`, got {function.__class__.__name__}; '
                f'{function!r}.'
            )
        
        code_object = function.__code__
        code_flags = code_object.co_flags
        if code_flags & CO_COROUTINE_ALL:
            return function
        
        if not code_flags & CO_GENERATOR:
            raise TypeError(
                f'`function` can only be given as generator or as coroutine type, got {function!r}, '
                f'co_flags={code_flags!r}.'
            )
        
        function.__code__ = type(code_object)(
            code_object.co_argcount,
            code_object.co_posonlyargcount,
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_flags | CO_ITERABLE_COROUTINE,
            code_object.co_code,
            code_object.co_consts,
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_qualname,
            code_object.co_firstlineno,
            code_object.co_linetable,
            code_object.co_exceptiontable,
            code_object.co_freevars,
            code_object.co_cellvars,
        )
        
        return function

elif sys.version_info >= (3, 8, 0):
    def to_coroutine(function):
        if not isinstance(function, FunctionType):
            raise TypeError(
                f'`function` can only be `{FunctionType.__name__}`, got {function.__class__.__name__}; '
                f'{function!r}.'
            )
        
        code_object = function.__code__
        code_flags = code_object.co_flags
        if code_flags & CO_COROUTINE_ALL:
            return function
        
        if not code_flags & CO_GENERATOR:
            raise TypeError(
                f'`function` can only be given as generator or as coroutine type, got {function!r}, '
                f'co_flags={code_flags!r}.'
            )
        
        function.__code__ = type(code_object)(
            code_object.co_argcount,
            code_object.co_posonlyargcount,
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_flags | CO_ITERABLE_COROUTINE,
            code_object.co_code,
            code_object.co_consts,
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_firstlineno,
            code_object.co_lnotab,
            code_object.co_freevars,
            code_object.co_cellvars,
        )
        
        return function
else:
    def to_coroutine(function):
        if not isinstance(function, FunctionType):
            raise TypeError(
                f'`function` can only be `{FunctionType.__name__}`, got {function.__class__.__name__}; '
                f'{function!r}.'
            )
        
        code_object = function.__code__
        code_flags = code_object.co_flags
        if code_flags & CO_COROUTINE_ALL:
            return function
        
        if not code_flags & CO_GENERATOR:
            raise TypeError(
                f'`function` can only be given as generator or as coroutine type, got {function!r}, '
                f'co_flags={code_flags!r}.'
            )
        
        function.__code__ = type(code_object)(
            code_object.co_argcount,
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_flags | CO_ITERABLE_COROUTINE,
            code_object.co_code,
            code_object.co_consts,
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_firstlineno,
            code_object.co_lnotab,
            code_object.co_freevars,
            code_object.co_cellvars,
        )
        
        return function

set_docs(to_coroutine,
    """
    Transforms the given generator function to coroutine function.
    
    Parameters
    ----------
    function : ``FunctionType``
        The generator function.
    
    Returns
    -------
    function : ``FunctionType``
    
    Raises
    ------
    TypeError
        - `function`'s type is incorrect.
        - `Function` cannot be turned to coroutine.
    """)



__all__ = ('SortedList', )

from .docs import has_docs
from .removed_descriptor import RemovedDescriptor


@has_docs
class SortedList(list):
    """
    An auto sorted list.
    
    Attributes
    ----------
    _reversed : `bool`
        Whether the list is reversed.
    """
    __slots__ = ('_reversed', )
    
    __setitem__ = RemovedDescriptor()
    insert = RemovedDescriptor()
    sort = RemovedDescriptor()
    __add__ = RemovedDescriptor()
    __radd__ = RemovedDescriptor()
    __iadd__ = RemovedDescriptor()
    __mul__ = RemovedDescriptor()
    __rmul__ = RemovedDescriptor()
    __imul__ = RemovedDescriptor()
    append = RemovedDescriptor()
    
    @has_docs
    def __init__(self, iterable=None, reverse=False):
        """
        Creates a new ``SortedList`` with the given parameters.
        
        Parameters
        ----------
        it : `None`, `iterable`, Optional
            An iterable to extend the created list with.
        reverse : `bool`, Optional
            Whether the created list should be reversed sorted.
        """
        self._reversed = reverse
        if (iterable is not None):
            self.extend(iterable)
            list.sort(self, reverse=reverse)
    
    
    @has_docs
    def __repr__(self):
        """Returns the sorted list's representation."""
        result = [self.__class__.__name__, '([']
        
        limit = len(self)
        if limit:
            index = 0
            while True:
                element = self[index]
                index +=1
                result.append(repr(element))
                
                if index == limit:
                    break
                
                result.append(', ')
                continue
        
        result.append('], reversed=')
        result.append(repr(self._reversed))
        result.append(')')
        
        return ''.join(result)
    
    def __getstate__(self):
        return self._reversed
    
    def __setstate__(self, state):
        self._reversed = state
    
    @property
    @has_docs
    def reverse(self):
        """
        A get-set descriptor to check or set how the sorted list sorted.
        """
        return self._reversed
    
    @reverse.setter
    def reverse(self, value):
        if self._reversed == value:
            return
        
        self._reversed = value
        list.reverse(self)
    
    @has_docs
    def add(self, value):
        """
        Adds a new value to the sorted list.
        
        Parameters
        ----------
        value : `Any`
            The value to insert to the SortedList.
        """
        index = self.relative_index(value)
        if index == len(self):
            # If the the index is at the end, then we just list append it.
            list.append(self, value)
            return
        
        element = self[index]
        if element == value:
            # If the element is same as the current, we overwrite it.
            list.__setitem__(self, index, value)
            return
        
        # No more special cases, simply list insert it
        list.insert(self, index, value)
        return
    
    
    @has_docs
    def remove(self, value):
        """
        Removes the given value from the SortedList.
        
        If the value is not in the list will not raise.
        
        Parameters
        ----------
        value : `Any`
            The value to remove.
        """
        index = self.relative_index(value)
        if index == len(self):
            # The element is not at self, leave
            return
        
        element = self[index]
        if element != value:
            # The element is different as the already added one att the correct position, leave.
            return
        
        # No more special case, remove it.
        list.__delitem__(self, index)
    
    
    @has_docs
    def extend(self, iterable):
        """
        Extends the SortedList with the given iterable object.
        
        Parameters
        ----------
        iterable : `iterable`
            Iterable object to extend the SortedList with.
        """
        ln = len(self)
        insert = list.insert
        bot = 0
        if self._reversed:
            if type(self) is not type(iterable):
                other = sorted(iterable, reverse=True)
            elif not iterable._reversed:
                other = reversed(iterable)
            else:
                other = iterable
            
            for value in other:
                top = ln
                while True:
                    if bot < top:
                        half = (bot + top) >> 1
                        if self[half] > value:
                            bot = half + 1
                        else:
                            top = half
                        continue
                    break
                insert(self, bot, value)
                ln += 1
        else:
            if type(self) is not type(iterable):
                other = sorted(iterable)
            elif iterable._reversed:
                other = reversed(iterable)
            else:
                other = iterable
            
            for value in other:
                top = ln
                while True:
                    if bot < top:
                        half = (bot + top) >> 1
                        if self[half] < value:
                            bot = half + 1
                        else:
                            top = half
                        continue
                    break
                insert(self, bot, value)
                ln += 1
    
    
    @has_docs
    def __contains__(self, value):
        """Returns whether the SortedList contains the given value."""
        index = self.relative_index(value)
        if index == len(self):
            return False
        
        if self[index] == value:
            return True
        
        return False
    
    
    @has_docs
    def index(self, value):
        """Returns the index of the given value inside of the SortedList."""
        index = self.relative_index(value)
        if index == len(self) or self[index] != value:
            raise ValueError(f'{value!r} is not in the {self.__class__.__name__}.')
        
        return index
    
    
    @has_docs
    def relative_index(self, value):
        """
        Returns the relative index of the given value if it would be inside of the SortedList.
        
        Parameters
        ----------
        value : `Any`
            The object's what's relative index is returned.
        
        Returns
        -------
        relative_index : `bool`
            The index where the given value would be inserted or should be inside of the SortedList.
        """
        bot = 0
        top = len(self)
        if self._reversed:
            while True:
                if bot < top:
                    half = (bot + top) >> 1
                    if self[half] > value:
                        bot = half + 1
                    else:
                        top = half
                    continue
                break
        else:
            while True:
                if bot < top:
                    half = (bot + top) >> 1
                    if self[half] < value:
                        bot = half + 1
                    else:
                        top = half
                    continue
                break
        
        return bot
    
    
    @has_docs
    def keyed_relative_index(self, value, key):
        """
        Returns the relative index of the given value if it would be inside of the SortedList.
        
        Parameters
        ----------
        value : `Any`
            The object's what's relative index is returned.
        key : `callable`
            A function that serves as a key for the sort comparison.
        
        Returns
        -------
        relative_index : `bool`
            The index where the given value would be inserted or should be inside of the SortedList.
        """
        bot = 0
        top = len(self)
        if self._reversed:
            while True:
                if bot < top:
                    half = (bot + top) >> 1
                    if key(self[half]) > value:
                        bot = half + 1
                    else:
                        top = half
                    continue
                break
        else:
            while True:
                if bot < top:
                    half = (bot + top) >> 1
                    if key(self[half]) < value:
                        bot = half + 1
                    else:
                        top = half
                    continue
                break
        return bot
    
    
    @has_docs
    def copy(self):
        """
        Copies the SortedList.
        
        Returns
        -------
        new : ``SortedList``
        """
        new = list.__new__(type(self))
        new._reversed = self._reversed
        list.extend(new, self)
        return new
    
    
    @has_docs
    def resort(self):
        """
        Resorts the SortedList.
        """
        list.sort(self, reverse=self._reversed)
    
    
    @has_docs
    def get(self, value, key, default=None):
        """
        Gets an element from the SortedList, what passed trough `key` equals to the given value.
        
        Parameters
        ----------
        value : `Any`
            The value to search in the SortedList.
        key : `callable`
            A function that serves as a key for the sort comparison.
        default : `Any` = `None`, Optional
            Default value to returns if no matching element was present.
        
        Returns
        -------
        element : `Any`, `default`
            The matched element or the `default` value if not found.
        """
        index = self.keyed_relative_index(value, key)
        if index == len(self):
            return default
        
        element = self[index]
        if key(element) == value:
            return element
        
        return default
    
    
    @has_docs
    def pop(self, value, key, default=None):
        """
        Gets and removes element from the SortedList, what's is passed trough `key` equals to the given value.
        
        Parameters
        ----------
        value : `Any`
            The value to search in the SortedList.
        key : `callable`
            A function that serves as a key for the sort comparison.
        default : `Any` = `None`, Optional
            Default value to returns if no matching element was present.
        
        Returns
        -------
        element : `Any`, `default`
            The matched element or the `default` value if not found.
        """
        index = self.keyed_relative_index(value, key)
        if index == len(self):
            return default
        
        element = self[index]
        if key(element) == value:
            del self[index]
            return element
        
        return default



__all__ = ('future_or_timeout', 'repeat_timeout',)

from threading import current_thread

from ...utils import include

from ..exceptions import CancelledError
from ..time import LOOP_TIME

from .handle_cancellers import _TimeoutHandleCanceller
from .task import Task


EventThread = include('EventThread')

def future_or_timeout(future, timeout):
    """
    If the given ``Future`` is not done till the given `timeout` occurs, set `TimeoutError` as it's exception.
    
    Parameters
    ----------
    future : ``Future``
        The future to set the timeout to.
    timeout : `float`
        The time after the given `future`'s exception is set as `TimeoutError`.
    
    Raises
    ------
    RuntimeError
        The future's event loop is already stopped.
    
    Notes
    -----
    For futures, which wait for more results and exceptions like ``Gatherer``, `TimeoutError` gives only 1 result,
    rigging it's results. This is not the case of ``FutureWM``, because that stops when the first exception is received.
    
    If `future_or_timeout` is used on ``WaitTillFirst``, ``WaitTillAll`` or on ``WaitTillAll``, then they stop
    collecting their result at the point, when the timeout occurs and they yield their actual result at that point
    without any specific exception.
    
    At the case of ``WaitContinuously``, when the timeout occurs the next yielded result will be `None` instead of a
    ``Future``.
    """
    loop = future._loop
    callback = _TimeoutHandleCanceller()
    handle = loop.call_later(timeout, callback, future)
    if handle is None:
        raise RuntimeError(
            f'`future_or_timeout` was called on future with a stopped loop; loop={loop!r}.'
        )
    
    callback._handle = handle
    future.add_done_callback(callback)


class repeat_timeout:
    """
    Repeats timeout on the code executed within a context of a for loop.
    
    The timeout it re-set with each iteration.
    
    The main difference between ``repeat_timeout` and ``future_or_timeout`` is, that this only renews the timeout if
    expired, saving a lot of resources on fast running loops compared to their timeout duration.
    
    Usage
    -----
    ```py
    try:
        with repeat_timeout(10.0) as loop:
            for _ in loop:
                await my_task(...)
    except TimeoutError:
        # my_task(...) did not complete within timeout.
    ```
    
    Attributes
    ----------
    _exception : `None`, ``CancelledError``
        The dropped exception to the task.
    _handle : `None`, ``TimerHandle``
        The handle to cancel when the loop is over.
    _last_set : `float`
        The last time when we repeated the timeout. Set as monotonic time. Defaults to `0.0` if was not set.
    _loop : ``EventThread``
        The current event loop.
    _task : ``Task``
        The currently running task.
    _timeout : `float`
        The time to drop `TimeoutError` after.
    """
    __slots__ = ('_exception', '_handle', '_last_set', '_loop', '_task', '_timeout')
    
    def __new__(cls, timeout):
        """
        Creates a new repeated timeouter.
        
        Parameters
        ----------
        timeout : `float`
            The time to drop `TimeoutException` after.
        """
        thread = current_thread()
        if not isinstance(thread, EventThread):
            raise RuntimeError(
                f'`repeat_timeout used outside of `{EventThread.__name__}`, at {thread!r}.'
            )
        
        task = thread.current_task
        if task is None:
            raise RuntimeError(
                f'`repeat_timeout` used outside of a `{Task.__name__}`.'
            )
        
        self = object.__new__(cls)
        self._last_set = 0.0
        self._loop = thread
        self._task = task
        self._timeout = timeout
        self._handle = thread.call_later(timeout, self)
        self._exception = None
        return self
    
    
    def __iter__(self):
        """Iterating a timeout handler returns itself."""
        return self
    
    
    def __next__(self):
        """Called when the timeout handler does a loop, resetting its timeout."""
        self._last_set = LOOP_TIME()
    
    
    def __enter__(self):
        """Entering a timeout handler returns itself."""
        return self
    
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exists the repeated timeout, dropping `TimeoutError` when cancelled."""
        handle = self._handle
        if (handle is not None):
            self._handle = None
            handle.cancel()
        
        exception = self._exception
        if (exception is not None) and (self._task._exception is exception):
            raise TimeoutError from None
        
        return False
    
    
    def __call__(self):
        """
        Checks whether the repeat timeout handler timed out.
        """
        last_set = self._last_set
        if last_set:
            self._last_set = 0.0
            handle = self._loop.call_at(last_set + self._timeout, self)
        else:
            handle = None
            exception = CancelledError()
            self._exception = exception
            self._task.set_exception_if_pending(exception)
        
        self._handle = handle



from .cycler import *
from .event_loop import *
from .event_loop_functionality_helpers import *
from .event_thread_suspender import *
from .event_thread_type import *
from .executor import *
from .handles import *
from .selector import *
from .server import *

__all__ = (
    *cycler.__all__,
    *event_loop.__all__,
    *event_loop_functionality_helpers.__all__,
    *event_thread_suspender.__all__,
    *event_thread_type.__all__,
    *executor.__all__,
    *handles.__all__,
    *selector.__all__,
    *server.__all__,
)



__all__ = ()

class Token:
    """
    Represents a token parsed by ``HighlightContextBase``.
    
    Attributes
    ----------
    type : `int`
        The token's identifier.
    value : `None`, `str`
        The token's value.
    """
    __slots__ = ('type', 'value',)
    
    def __new__(cls, type_, value):
        """
        Creates a new ``Token``.
        
        Parameters
        ----------
        type_ : `int`
            The token's identifier.
        value : `None`, `str`
            The token's value.
        """
        self = object.__new__(cls)
        self.type = type_
        self.value = value
        return self
    
    def __repr__(self):
        """Returns the token's representation."""
        return f'{self.__class__.__name__}({self.type}, {self.value!r})'


def _merge_tokens(tokens, start_index, end_index):
    """
    Merges the tokens inside of the given range.
    
    Parameters
    ----------
    tokens : `list` of ``Token``
        The tokens, which will have it's slice merged.
    start_index : `int`
        The first token's index to merge.
    end_index : `int`
        The last token's index +1 to merge.
    """
    values = []
    
    for token_index in range(start_index, end_index):
        value = tokens[token_index].value
        if (value is not None) and value:
            values.append(value)
    
    value = ''.join(values)
    
    del tokens[start_index + 1 : end_index]
    tokens[start_index].value = value
